// Test Align for updating !!!!!!!!!!!!!!!!!!

#include "StdAfx.h"

#include <stdio.h>

#include "Common/Buffer.h"
#include "Common/StringConvert.h"
#include "Common/WildCard.h"

#include "Windows/COM.h"
#include "Windows/FileFind.h"
#include "Windows/FileIO.h"
#include "Windows/FileDir.h"

#include "../../../7zip/UI/Far/FarUtils.h"

#include "DocFileUtils.h"
#include "Messages.h"
#include "StorageWalk.h"

using namespace NWindows;
using namespace NFar;

static const int kDescriptionMaxSize = 256;
static const char *kCommandPrefix = "DocFile";
static const char *kHelpTopicExtrFromDocFile =  "ExtractDocFile";
static const char *kHelpTopicConfig =  "DocFileConfig";
static const char *kPliginNameForRegestry = "DocFile";
static const char *kRegisrtryMainKeyName = "";
static const char *kRegisrtryValueNameEnabled = "Enabled";

static bool kPluginEnabledDefault = true;

HINSTANCE g_hInstance;
#ifndef _UNICODE
bool g_IsNT = false;
static bool IsItWindowsNT()
{
  OSVERSIONINFO versionInfo;
  versionInfo.dwOSVersionInfoSize = sizeof(versionInfo);
  if (!::GetVersionEx(&versionInfo))
    return false;
  return (versionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT);
}
#endif

BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID)
{
  if (dwReason == DLL_PROCESS_ATTACH)
  {
    g_hInstance = hInstance;
    #ifndef _UNICODE
    g_IsNT = IsItWindowsNT();
    #endif
  }
  return TRUE;
}

struct COptions
{
  bool Enabled;
};

static COptions g_Options;

EXTERN_C void WINAPI SetStartupInfo(struct PluginStartupInfo *info)
{
  MY_TRY_BEGIN;
  g_StartupInfo.Init(*info, kPliginNameForRegestry);
  g_Options.Enabled = g_StartupInfo.QueryRegKeyValue(
      HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameEnabled, kPluginEnabledDefault);
  MY_TRY_END1("SetStartupInfo");
}

static UString ConvertNameTo_FarName(const UString &name)
{
  UString tmp;
  if (CompoundMsiNameToFileName(name, tmp))
    return tmp;
  return CompoundNameToFileName(name, false);
}

static UString ConvertNameTo_FileSysName(const UString &name)
{
  UString tmp;
  if (CompoundMsiNameToFileName(name, tmp))
    return tmp;
  return CompoundNameToFileName(name, true);
}

static UString Convert_FileSysName_To_ComName(const UString &name)
{
  UString tmp;
  if (FileNameToCompoundMsiName(name, tmp))
    return tmp;
  return FileNameToCompoundName(name);
}

static UString OpenStorages_GetCurDirPath(const COpenStorages &openStorages)
{
  UString result;
  for (int i = 1; i < openStorages.m_Storages.Size(); i++)
  {
    if (i > 1)
      result += WCHAR_PATH_SEPARATOR;
    result += ConvertNameTo_FarName(openStorages.m_Names[i]);
  }
  return result;
}

class CPlugin
{
  NCOM::CComInitializer m_ComInitializer;
  CByteBuffer m_Buffer;
  CObjectVector<CItemStat> m_Items;
  UString m_CurrentDir;
  CSysString m_PannelTitle;
  
  InfoPanelLine m_InfoLines[1];

  char m_FileNameBuffer[1024];
  char m_CurrentDirBuffer[1024];
  char m_PannelTitleBuffer[1024];

  NFileOperationReturnCode::EEnum GetOneFile(const UString &destPath, const UString &name);
  NFileOperationReturnCode::EEnum GetOneItem(const UString &destPath, const CItemStat &itemStat);
  NFileOperationReturnCode::EEnum CPlugin::PutOneItem(const UString &pathPrefix, const NFile::NFind::CFileInfoW &fi);
  NFileOperationReturnCode::EEnum PutPanelItems(struct PluginPanelItem *panelItems, int itemsNumber);

  void EnterToDirectory(const UString &dirName);

  void LoadItems(CObjectVector<CItemStat> &items);
  void LoadMainItems();

public:
  CSysString m_FileName;
  COpenStorages m_OpenStorages;

  int GetFindData(PluginPanelItem **pPanelItem,int *itemsNumber, int OpMode);
  void FreeFindData(PluginPanelItem *PanelItem,int ItemsNumber);
  int SetDirectory(char *path, int opMode);
  void GetOpenPluginInfo(struct OpenPluginInfo *info);

  int DeleteFiles(PluginPanelItem *panelItems, int itemsNumber, int opMode);
  NFileOperationReturnCode::EEnum GetFiles(struct PluginPanelItem *pi, int itemsNumber,
      int moveMode, char *destPath, int opMode);

  NFileOperationReturnCode::EEnum PutFiles(struct PluginPanelItem *panelItems, int itemsNumber,
      int moveMode, int opMode);

  int ProcessKey(int key, unsigned int controlState);
};


static HANDLE MyOpenFilePlugin(const char *name)
{
  CSysString normalizedName = name;
  normalizedName.Trim();
  CSysString fullPath;
  NFile::NDirectory::MyGetFullPathName(normalizedName, fullPath);
  NFile::NFind::CFileInfo fi;
  if (!fi.Find(fullPath))
    return INVALID_HANDLE_VALUE;
  CMyComPtr<IStorage> newStorage;
  if (OpenRootStorage(MultiByteToUnicodeString(fullPath, CP_OEMCP), &newStorage, fi.IsReadOnly()) != S_OK)
    return INVALID_HANDLE_VALUE;
  CPlugin *plugin = new CPlugin();
  if (plugin == NULL)
    return(INVALID_HANDLE_VALUE);
  plugin->m_OpenStorages.m_ReadOnly = fi.IsReadOnly();
  plugin->m_OpenStorages.AddRootStorage(newStorage);
  plugin->m_FileName = fullPath;
  return (HANDLE)plugin;
}

EXTERN_C HANDLE WINAPI OpenFilePlugin(char *name, const unsigned char * /* data */, unsigned int /* dataSize */)
{
  MY_TRY_BEGIN;
  NCOM::CComInitializer comInitializer;
  if (name == NULL || (!g_Options.Enabled))
  {
    // if (!Opt.ProcessShiftF1)
      return(INVALID_HANDLE_VALUE);
  }
  return MyOpenFilePlugin(name);
  MY_TRY_END2("OpenFilePlugin", INVALID_HANDLE_VALUE);
}

EXTERN_C HANDLE WINAPI OpenPlugin(int openFrom, int item)
{
  MY_TRY_BEGIN;
  NCOM::CComInitializer comInitializer;
  if (openFrom == OPEN_COMMANDLINE)
  {
    CSysString fileName = (const char *)item;
    if (fileName.IsEmpty())
      return INVALID_HANDLE_VALUE;
    if (fileName.Length() >= 2 && fileName[0] == '\"' && fileName[fileName.Length() - 1] == '\"')
      fileName = fileName.Mid(1, fileName.Length() - 2);

    return MyOpenFilePlugin(fileName);
  }
  if (openFrom == OPEN_PLUGINSMENU)
  {
    PluginPanelItem pluginPanelItem;
    if (!g_StartupInfo.ControlGetActivePanelCurrentItemInfo(pluginPanelItem))
      throw 142134;
    return MyOpenFilePlugin(pluginPanelItem.FindData.cFileName);
  }
  return INVALID_HANDLE_VALUE;
  MY_TRY_END2("OpenPlugin", INVALID_HANDLE_VALUE);
}

EXTERN_C void WINAPI ClosePlugin(HANDLE hPlugin)
{
  MY_TRY_BEGIN;
  delete (CPlugin *)hPlugin;
  MY_TRY_END1("ClosePlugin");
}

EXTERN_C int WINAPI GetFindData(HANDLE _aPlugin, struct PluginPanelItem **panelItems, int *itemsNumber,int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)_aPlugin;
  return(plugin->GetFindData(panelItems, itemsNumber, opMode));
  MY_TRY_END2("GetFindData", FALSE);
}

void CPlugin::LoadItems(CObjectVector<CItemStat> &items)
{
  CMyComPtr<IEnumSTATSTG> enumSTATSTG;
  HRESULT result = m_OpenStorages.m_Storages.Back()->EnumElements(0, NULL, 0, &enumSTATSTG);
  if (result != S_OK)
    throw "Can not enumerate storage elements";
  STATSTG stat;
  items.Clear();
  for (;;)
  {
    HRESULT hResult = enumSTATSTG->Next(1, &stat, NULL);
    if (hResult != S_OK) // S_FALSE is possible?
      break;
    CItemStat itemStat;
    itemStat.SetFrom(stat);
    items.Add(itemStat);
    CoTaskMemFree(stat.pwcsName);
  }
}

void CPlugin::LoadMainItems()
{
  LoadItems(m_Items);
}

/*
AString DescriptonFromSTGTY(STGTY stgty)
{
  switch (stgty)
  {
    case STGTY_STORAGE: return "Storage";
    case STGTY_STREAM: return "Stream";
    case STGTY_LOCKBYTE: return "Byte Array";
    case STGTY_PROPERTY: return "Property Storage";
    default: throw 40351;
  }
}
*/

int CPlugin::GetFindData(PluginPanelItem **panelItems, int *itemsNumber, int opMode)
{
  CScreenRestorer screenRestorer;
  CRecordVector<PluginPanelItem> pluginPanelItems;
  if ((opMode & OPM_SILENT) == 0 && (opMode & OPM_FIND ) == 0)
  {
    screenRestorer.Save();
    const char *msgItems[]=
    {
      g_StartupInfo.GetMsgString(NMessageID::kArcReadTitle),
      g_StartupInfo.GetMsgString(NMessageID::kArcReading)
    };
    g_StartupInfo.ShowMessage(0, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 0);
  }
  LoadMainItems();

  for (int i = 0; i < m_Items.Size(); i++)
  {
    const CItemStat &item = m_Items[i];
    
    PluginPanelItem pi;
    pi.FindData.dwFileAttributes = 0; // change it;
    if (item.IsDir())
      pi.FindData.dwFileAttributes |= FILE_ATTRIBUTE_DIRECTORY; // change it;

    strcpy(pi.FindData.cFileName, GetOemString(ConvertNameTo_FarName(item.Name)));
    pi.FindData.cAlternateFileName[0] = 0;

    FILETIME ft;
    ft.dwHighDateTime = 0;
    ft.dwLowDateTime = 0;

    if (item.IsDir())
    {
      pi.FindData.ftCreationTime = item.CTime;
      pi.FindData.ftLastAccessTime = item.ATime;
      pi.FindData.ftLastWriteTime = item.MTime;
      pi.FindData.nFileSizeHigh = 0;
      pi.FindData.nFileSizeLow = 0;
    }
    else
    {
      pi.FindData.ftCreationTime = ft;
      pi.FindData.ftLastAccessTime = ft;
      pi.FindData.ftLastWriteTime = ft;
      pi.FindData.nFileSizeHigh = item.Size.HighPart;
      pi.FindData.nFileSizeLow = item.Size.LowPart;
    }
    pi.PackSizeHigh = pi.FindData.nFileSizeHigh;
    pi.PackSize = pi.FindData.nFileSizeLow;
    pi.Flags = 0;
    pi.NumberOfLinks = 0;

    /*
    const CSysString &description = DescriptonFromSTGTY(item.Type);
    const CSysString &description = BSTRToOemString(ClassIDToString(item.ClassID));
    if (description.IsEmpty())
      pi.Description = NULL;
    else
    {
      pi.Description = new char[description.Length() + 1];
      strcpy(pi.Description, description.GetPointer());
    }
    */
    pi.Description = NULL;

    pi.Owner = NULL;
    pi.CustomColumnData = NULL;
    pi.CustomColumnNumber = 0;
    pi.UserData = i;
    pluginPanelItems.Add(pi);
  }
  int numItems = pluginPanelItems.Size();
  *panelItems = new PluginPanelItem[numItems];
  memcpy(*panelItems, &pluginPanelItems.Front(), sizeof(PluginPanelItem) * numItems);
  *itemsNumber = numItems;
  return TRUE;
}

EXTERN_C void WINAPI FreeFindData(HANDLE _aPlugin, struct PluginPanelItem *panelItems, int itemsNumber)
{
  // MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)_aPlugin;
  plugin->FreeFindData(panelItems, itemsNumber);
  // MY_TRY_END1("FreeFindData");
}

void CPlugin::FreeFindData(struct PluginPanelItem *panelItems, int itemsNumber)
{
  for (int i = 0; i < itemsNumber; i++)
    if (panelItems[i].Description != NULL)
      delete []panelItems[i].Description;
  delete []panelItems;
}

EXTERN_C int WINAPI GetFiles(HANDLE hPlugin, struct PluginPanelItem *panelItems,
    int itemsNumber, int moveMode, char *destPath, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->GetFiles(panelItems, itemsNumber, moveMode, destPath, opMode));
  MY_TRY_END2("GetFiles", NFileOperationReturnCode::kError);
}

static const DWORD kBufferSize = 0x40000;

NFileOperationReturnCode::EEnum CPlugin::GetOneFile(const UString &destPath, const UString &name)
{
  CMyComPtr<IStream> stream;
  m_OpenStorages.OpenStream(name, &stream);

  m_Buffer.SetCapacity(kBufferSize);
 
  NFile::NIO::COutFile file;
  if (!file.Create(destPath + ConvertNameTo_FileSysName(name), true))
    throw "file create error";

  UInt64 streamSize = GetStreamSize(stream);
  while (streamSize > 0)
  {
    ULONG processed;
    ULONG curSize = (ULONG)MyMin(streamSize, (UInt64)kBufferSize);
    if (stream->Read((BYTE *)m_Buffer, curSize, &processed) != S_OK || processed != curSize)
      throw "Stream read error";
    UInt32 written;
    if (!file.Write(m_Buffer, curSize, written) || written != curSize)
      throw "Write to file error";
    streamSize -= curSize;
  }
  return NFileOperationReturnCode::kSuccess;
}

class CCurrentStorageRestorer
{
  COpenStorages *m_OpenStorages;
  int m_Index;
public:
  CCurrentStorageRestorer(COpenStorages &openStorages): m_OpenStorages(&openStorages)
      { m_Index = m_OpenStorages->m_Storages.Size(); }
  ~CCurrentStorageRestorer() { Restore(); }
  void Restore() { m_OpenStorages->CloseAllUpToSpecified(m_Index); }
};

NFileOperationReturnCode::EEnum CPlugin::GetOneItem(const UString &destPath, const CItemStat &itemStat)
{
  const UString &name = itemStat.Name;
  if (itemStat.IsDir())
  {
    CCurrentStorageRestorer currentStorageRestorer(m_OpenStorages);
    m_OpenStorages.OpenNewStorage(name);
    UString newPath = destPath + ConvertNameTo_FileSysName(name) + WCHAR_PATH_SEPARATOR;
    NFile::NDirectory::CreateComplexDirectory(newPath); // can be optimized
    CObjectVector<CItemStat> items;
    LoadItems(items);
    for (int i = 0; i < items.Size(); i++)
    {
      NFileOperationReturnCode::EEnum result = GetOneItem(newPath, items[i]);
      if (result != NFileOperationReturnCode::kSuccess)
        return result;
    }
    return NFileOperationReturnCode::kSuccess;
  }
  else
    return GetOneFile(destPath, name);
}

NFileOperationReturnCode::EEnum CPlugin::GetFiles(struct PluginPanelItem *panelItems,
    int itemsNumber, int moveMode, char *_aDestPath, int opMode)
{
  CScreenRestorer screenRestorer;
  if ((opMode & OPM_SILENT) == 0 && (opMode & OPM_FIND ) == 0)
  {
    screenRestorer.Save();
    const char *msgItems[]=
    {
      g_StartupInfo.GetMsgString(NMessageID::kArcReadTitle),
      g_StartupInfo.GetMsgString(NMessageID::kArcGetting)
    };
    g_StartupInfo.ShowMessage(0, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 0);
  }

  CSysString destPath = _aDestPath;
  NFile::NName::NormalizeDirPathPrefix(destPath);

  if ((opMode & OPM_SILENT) == 0)
  {
    const int kPathIndex = 2;
    const int kOkButtonIndex = 4;
    for (;;)
    {
      struct CInitDialogItem initItems[]={
        { DI_DOUBLEBOX, 3, 1, 72, 6, false, false, 0, false, NMessageID::kExtractTitle, NULL, NULL },
        { DI_TEXT, 5, 2, 0, 0, false, false, 0, false, NMessageID::kExtractTo, NULL, NULL },
        
        //{ DI_EDIT, 5, 3, 70, 3, true, false, DIF_HISTORY, false, -1, destPath.GetPointer(), kExractPathHistoryName},
        { DI_EDIT, 5, 3, 70, 3, true, false, 0, false, -1, destPath, NULL},
        
        { DI_TEXT, 3, 4, 0, 0, false, false, DIF_BOXCOLOR|DIF_SEPARATOR, false, -1, "", NULL  },
        { DI_BUTTON, 0, 5, 0, 0, false, false, DIF_CENTERGROUP, true, NMessageID::kExtrExtract, NULL, NULL  },
        { DI_BUTTON, 0, 5, 0, 0, false, false, DIF_CENTERGROUP, false, NMessageID::kExtrCancel, NULL, NULL  }
      };
      
      const int kNumDialogItems = sizeof(initItems) / sizeof(initItems[0]);
      FarDialogItem dialogItems[kNumDialogItems];
      g_StartupInfo.InitDialogItems(initItems, dialogItems, kNumDialogItems);
      
      int askCode = g_StartupInfo.ShowDialog(76, 8, kHelpTopicExtrFromDocFile, dialogItems, kNumDialogItems);
      if (askCode != kOkButtonIndex)
        return NFileOperationReturnCode::kInterruptedByUser;
      
      destPath = dialogItems[kPathIndex].Data;
      destPath.Trim();
      if (!destPath.IsEmpty() && destPath[destPath.Length() - 1] == WCHAR_PATH_SEPARATOR)
        break;
      g_StartupInfo.ShowMessage("You must specify directory path");
    }
  }

  NFile::NDirectory::CreateComplexDirectory(destPath);
  UString destPath2 = GetUnicodeString(destPath, CP_OEMCP);

  for (int i = 0; i < itemsNumber; i++)
  {
    int index = panelItems[i].UserData;
    NFileOperationReturnCode::EEnum result = GetOneItem(destPath2, m_Items[index]);
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  if (moveMode != 0)
  {
    if (DeleteFiles(panelItems, itemsNumber, opMode) == FALSE)
      return NFileOperationReturnCode::kError;
  }
  return NFileOperationReturnCode::kSuccess;
}

EXTERN_C int WINAPI SetDirectory(HANDLE hPlugin, char *path, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->SetDirectory(path, opMode));
  MY_TRY_END2("SetDirectory", FALSE);
}

void CPlugin::EnterToDirectory(const UString &dirName)
{
  LoadMainItems();
  for (int i = 0; i < m_Items.Size(); i++)
  {
    const CItemStat &item = m_Items[i];
    if (!item.IsDir())
      continue;
    if (dirName.CompareNoCase(ConvertNameTo_FarName(item.Name)) == 0)
    {
      m_OpenStorages.OpenNewStorage(item.Name);
      return;
    }
  }
  throw "can't find directory";
}

int CPlugin::SetDirectory(char *aszDir, int /* opMode */)
{
  UString path = GetUnicodeString(aszDir, CP_OEMCP);
  if (path == L"\\")
    m_OpenStorages.CloseAllExceptRoot();
  else if (path == L"..")
  {
    if (m_OpenStorages.m_Storages.Size() <= 1)
      throw 40312;
    m_OpenStorages.CloseLastStorage();
  }
  else
  {
    UStringVector pathParts;
    SplitPathToParts(path, pathParts);
    for (int i = 0; i < pathParts.Size(); i++)
      EnterToDirectory(pathParts[i]);
  }
  m_CurrentDir = OpenStorages_GetCurDirPath(m_OpenStorages);
  return TRUE;
}

EXTERN_C void WINAPI GetPluginInfo(struct PluginInfo *info)
{
  MY_TRY_BEGIN;
  info->StructSize = sizeof(*info);
  info->Flags = 0;
  info->DiskMenuStrings = NULL;
  info->DiskMenuNumbers = NULL;
  info->DiskMenuStringsNumber = 0;
  static const char *pluginMenuStrings[1];
  pluginMenuStrings[0] = g_StartupInfo.GetMsgString(NMessageID::kCfgLine);
  info->PluginMenuStrings = (char **)pluginMenuStrings;
  info->PluginMenuStringsNumber = 1;
  static const char *PluginCfgStrings[1];
  PluginCfgStrings[0] = g_StartupInfo.GetMsgString(NMessageID::kCfgLine);
  info->PluginConfigStrings = (char **)PluginCfgStrings;
  info->PluginConfigStringsNumber = sizeof(PluginCfgStrings) / sizeof(PluginCfgStrings[0]);
  info->CommandPrefix = (char *)kCommandPrefix;
  MY_TRY_END1("GetPluginInfo");
}

EXTERN_C int WINAPI Configure(int /* itemNumber */)
{
  MY_TRY_BEGIN;

  const int kEnabledCheckBoxIndex = 1;
  const int kYSize = 7;

  struct CInitDialogItem initItems[]=
  {
    { DI_DOUBLEBOX, 3, 1, 72, kYSize - 2, false, false, 0, false, NMessageID::kConfigTitle, NULL, NULL },
    { DI_CHECKBOX, 5, 2, 0, 0, true, g_Options.Enabled, 0, false, NMessageID::kConfigPluginEnabled, NULL, NULL },
    { DI_TEXT, 5, 3, 0, 0, false, false, DIF_BOXCOLOR | DIF_SEPARATOR, false, -1, "", NULL },
    { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, true, NMessageID::kOk, NULL, NULL },
    { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, false, NMessageID::kCancel, NULL, NULL },
  };

  const int kNumDialogItems = sizeof(initItems) / sizeof(initItems[0]);
  const int kOkButtonIndex = kNumDialogItems - 2;

  FarDialogItem dialogItems[kNumDialogItems];
  g_StartupInfo.InitDialogItems(initItems, dialogItems, kNumDialogItems);

  int askCode = g_StartupInfo.ShowDialog(76, kYSize, kHelpTopicConfig, dialogItems, kNumDialogItems);
  if (askCode != kOkButtonIndex)
    return FALSE;

  g_Options.Enabled = BOOLToBool(dialogItems[kEnabledCheckBoxIndex].Selected);

  g_StartupInfo.SetRegKeyValue(HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameEnabled, g_Options.Enabled);
  return TRUE;
  MY_TRY_END2("Configure", FALSE);
}

EXTERN_C void WINAPI GetOpenPluginInfo(HANDLE hPlugin,struct OpenPluginInfo *info)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  plugin->GetOpenPluginInfo(info);
  MY_TRY_END1("GetOpenPluginInfo");
}

static char *kPluginFormatName = "DocFile";

static void CopyStrLimited(char *dest, const AString &src, int len)
{
  len--;
  if (src.Length() < len)
    len = src.Length();
  memcpy(dest, src, sizeof(dest[0]) * len);
  dest[len] = 0;
}

#define COPY_STR_LIMITED(dest, src) CopyStrLimited(dest, src, sizeof(dest) / sizeof(dest[0]))

void CPlugin::GetOpenPluginInfo(struct OpenPluginInfo *info)
{
  info->StructSize = sizeof(*info);
  info->Flags = OPIF_USEFILTER | OPIF_USESORTGROUPS | OPIF_USEHIGHLIGHTING | OPIF_ADDDOTS | OPIF_COMPAREFATTIME;

  COPY_STR_LIMITED(m_FileNameBuffer, m_FileName);
  info->HostFile = m_FileNameBuffer; // test it it is not static
  
  COPY_STR_LIMITED(m_CurrentDirBuffer, GetOemString(m_CurrentDir));
  info->CurDir = m_CurrentDirBuffer;

  info->Format = kPluginFormatName;

  CSysString name;
  {
    CSysString fullPath;
    int index;
    NFile::NDirectory::MyGetFullPathName(m_FileName, fullPath, index);
    name = fullPath.Mid(index);
  }
  m_PannelTitle = "DocFile:" + name;
  if (!m_CurrentDir.IsEmpty())
  {
    m_PannelTitle += '\\';
    m_PannelTitle += GetOemString(m_CurrentDir);
  }
 
  COPY_STR_LIMITED(m_PannelTitleBuffer, m_PannelTitle);
  info->PanelTitle = m_PannelTitleBuffer;

  strcpy(m_InfoLines[0].Text, "DocFile Type File");
  strcpy(m_InfoLines[0].Data, "Normal DOC XLS");
  m_InfoLines[0].Separator = 0;

  info->InfoLines = m_InfoLines;
  info->InfoLinesNumber = sizeof(m_InfoLines) / sizeof(m_InfoLines[0]);

  info->DescrFiles = NULL;
  info->DescrFilesNumber = 0;

  info->PanelModesArray = NULL;
  info->PanelModesNumber = 0;

  info->StartPanelMode = 0;
  info->StartSortMode = 0;
  info->KeyBar = NULL;
  info->ShortcutData = NULL;
}


static const char *kFileMustBeLess4GBMessage = "File size must be less than 4GB";

EXTERN_C int WINAPI PutFiles(HANDLE hPlugin, struct PluginPanelItem *panelItems,
    int itemsNumber, int moveMode, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->PutFiles(panelItems, itemsNumber, moveMode, opMode));
  MY_TRY_END2("PutFiles", NFileOperationReturnCode::kError);
}

NFileOperationReturnCode::EEnum CPlugin::PutOneItem(const UString &pathPrefix,
    const NFile::NFind::CFileInfoW &fi)
{
  const UString &path = pathPrefix + fi.Name;
  UString name = Convert_FileSysName_To_ComName(fi.Name);
  if (fi.IsDir())
  {
    CCurrentStorageRestorer currentStorageRestorer(m_OpenStorages);
    m_OpenStorages.OpenStorageAlways(name);

    NFile::NFind::CEnumeratorW enumerator(path + L"\\*");
    NFile::NFind::CFileInfoW fiNew;
    while (enumerator.Next(fiNew))
    {
      NFileOperationReturnCode::EEnum result = PutOneItem(path + L"\\", fiNew);
      if (result != NFileOperationReturnCode::kSuccess)
        return result;
    }
  }
  else
  {
    if (fi.Size >= ((UInt64)1 << 32))
      throw kFileMustBeLess4GBMessage;
    CMyComPtr<IStream> stream;
    m_OpenStorages.CreateStream(name, &stream);
    m_Buffer.SetCapacity(kBufferSize);
    NFile::NIO::CInFile file;
    if (!file.Open(path))
      throw "file open Error";
    UInt64 streamSize;
    file.GetLength(streamSize);
    while (streamSize > 0)
    {
      ULONG written;
      UInt32 curSize = (UInt32)MyMin(streamSize, (UInt64)kBufferSize);
      UInt32 processed;
      if (!file.Read(m_Buffer, curSize, processed) || processed != curSize)
        throw "Read from file error";
      if (stream->Write((BYTE *)m_Buffer, curSize, &written) != S_OK || written != curSize)
        throw "Write error";
      streamSize -= curSize;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::PutFiles(struct PluginPanelItem *panelItems, int itemsNumber,
    int moveMode, int /* opMode */ )
{
  for (int i = 0; i < itemsNumber; i++)
  {
    const PluginPanelItem &pi = panelItems[i];
    const FarFindData &fd = pi.FindData;
    NFile::NFind::CFileInfoW fi;
    fi.Attrib = fd.dwFileAttributes;
    fi.CTime = fd.ftCreationTime;
    fi.ATime = fd.ftLastAccessTime;
    fi.MTime = fd.ftLastWriteTime;
    fi.Size = fd.nFileSizeLow | ((UInt64)fd.nFileSizeHigh << 32);
    fi.Name = GetUnicodeString(fd.cFileName, CP_OEMCP);
    NFileOperationReturnCode::EEnum result = PutOneItem(L"", fi);
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  if (moveMode != 0)
  {
    for (int i = 0; i < itemsNumber; i++)
    {
      const PluginPanelItem &pluginPanelItem = panelItems[i];
      bool result;
      if (NFile::NFind::NAttributes::IsDir(pluginPanelItem.FindData.dwFileAttributes))
        result = NFile::NDirectory::RemoveDirectoryWithSubItems(pluginPanelItem.FindData.cFileName);
      else
        result = NFile::NDirectory::DeleteFileAlways(pluginPanelItem.FindData.cFileName);
      if (!result)
        return NFileOperationReturnCode::kError;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}

EXTERN_C int WINAPI DeleteFiles(HANDLE hPlugin, PluginPanelItem *panelItems, int itemsNumber, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)hPlugin;
  return(plugin->DeleteFiles(panelItems , itemsNumber, opMode));
  MY_TRY_END2("DeleteFiles", FALSE);
}

int CPlugin::DeleteFiles(PluginPanelItem *panelItems, int itemsNumber, int opMode)
{
  if (itemsNumber == 0)
    return FALSE;
  if ((opMode & OPM_SILENT) == 0)
  {
    const char *msgItems[]=
    {
      g_StartupInfo.GetMsgString(NMessageID::kDeleteTitle),
      g_StartupInfo.GetMsgString(NMessageID::kDeleteFiles),
      g_StartupInfo.GetMsgString(NMessageID::kDeleteDelete),
      g_StartupInfo.GetMsgString(NMessageID::kDeleteCancel)
    };
    char msg[1024];
    if (itemsNumber == 1)
    {
      sprintf(msg, g_StartupInfo.GetMsgString(NMessageID::kDeleteFile), panelItems[0].FindData.cFileName);
      msgItems[1] = msg;
    }
    else if (itemsNumber > 1)
    {
      sprintf(msg, g_StartupInfo.GetMsgString(NMessageID::kDeleteNumberOfFiles), itemsNumber);
      msgItems[1] = msg;
    }
    if (g_StartupInfo.ShowMessage(FMSG_WARNING, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 2) != 0)
      return FALSE;
  }

  for (int i = 0; i < itemsNumber; i++)
  {
    int index = panelItems[i].UserData;
    m_OpenStorages.DestroyElement(m_Items[index].Name);
  }
  return TRUE;
}

EXTERN_C int WINAPI ProcessKey(HANDLE hPlugin, int key, unsigned int controlState)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return (plugin->ProcessKey(key, controlState));
  MY_TRY_END2("ProcessKey", FALSE);
}

int CPlugin::ProcessKey(int key, unsigned int controlState)
{
  if (controlState != 0)
    return FALSE;
  if (key != VK_F7)
    return FALSE;
  CSysString subStorageName;
  for (;;)
  {
    const int kPathIndex = 2;
    const int kOkButtonIndex = 4;
    const int kYSize = 8;
    struct CInitDialogItem initItems[]= {
      { DI_DOUBLEBOX, 3, 1, 72, kYSize - 2, false, false, 0, false, NMessageID::kConfigPluginCreateSubStorageTitle, NULL, NULL },
      
      //{ DI_EDIT, 5, 3, 70, 3, true, false, DIF_HISTORY, false, -1, destPath.GetPointer(), kExractPathHistoryName},
     { DI_TEXT, 5, 2, 0, 0, false, false, 0, false, NMessageID::kConfigPluginCreateSubStorageLabel, NULL, NULL },
     { DI_EDIT, 5, 3, 70, 3, true, false, 0, false, -1, subStorageName, NULL},
      
      { DI_TEXT, 3, kYSize - 4, 0, 0, false, false, DIF_BOXCOLOR|DIF_SEPARATOR, false, -1, "", NULL  },
      { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, true, NMessageID::kOk, NULL, NULL  },
      { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, false, NMessageID::kCancel, NULL, NULL  }
    };
    
    const int kNumDialogItems = sizeof(initItems) / sizeof(initItems[0]);
    FarDialogItem dialogItems[kNumDialogItems];
    g_StartupInfo.InitDialogItems(initItems, dialogItems, kNumDialogItems);
    
    int askCode = g_StartupInfo.ShowDialog(76, kYSize, NULL, dialogItems, kNumDialogItems);
    if (askCode != kOkButtonIndex)
      return FALSE;
    
    subStorageName = dialogItems[kPathIndex].Data;
    subStorageName.Trim();

    if (m_OpenStorages.CreateStorage(MultiByteToUnicodeString(subStorageName, CP_OEMCP)) == S_OK)
      break;
    g_StartupInfo.ShowMessage(NMessageID::kConfigPluginCreateSubStorageAlreadyExist);
  }
  g_StartupInfo.Control(this, FCTL_UPDATEPANEL, (void *)1);
  g_StartupInfo.Control(this, FCTL_REDRAWPANEL, NULL);
  PanelInfo panelInfo;
  if (!g_StartupInfo.ControlGetActivePanelInfo(panelInfo))
    return 242307;
  for (int i = 0; i < panelInfo.ItemsNumber; i++)
  {
    const PluginPanelItem &pi = panelInfo.PanelItems[i];
    if (subStorageName.Compare(pi.FindData.cFileName) == 0)
    {
      PanelRedrawInfo panelRedrawInfo;
      panelRedrawInfo.CurrentItem = i;
      panelRedrawInfo.TopPanelItem = 0;
      g_StartupInfo.Control(this, FCTL_REDRAWPANEL, &panelRedrawInfo);
      break;
    }
  }
  return TRUE;
}
