// StorageWalk.h

#ifndef __STORAGE_WALK_H
#define __STORAGE_WALK_H

#include "Common/MyCom.h"
#include "Common/MyString.h"
#include "Common/Types.h"

struct CItemStat
{
  UString Name;
  STGTY Type;
  ULARGE_INTEGER Size;
  FILETIME MTime;
  FILETIME CTime;
  FILETIME ATime;
  DWORD grfMode;
  DWORD grfLocksSupported;
  CLSID ClassID;
  DWORD grfStateBits;
  
  void SetFrom(const STATSTG &aSTATSTG);
  bool IsDir() const { return (Type == STGTY_STORAGE); };
};

class COpenStorages
{
  DWORD GetAccessMode() const { return m_ReadOnly ? STGM_READ: STGM_READWRITE; }
  DWORD GetAccessMode2() const { return GetAccessMode() | STGM_DIRECT | STGM_SHARE_EXCLUSIVE; }
public:
  CObjectVector<CMyComPtr<IStorage> > m_Storages;
  CObjectVector<UString> m_Names;
  bool m_ReadOnly;
  
  ~COpenStorages() { CloseAllStorages(); }
  void AddRootStorage(IStorage *newStorage);
  HRESULT CreateStorage(const WCHAR *name);
  void OpenNewStorage(const WCHAR *name);
  void OpenStream(const WCHAR *name, IStream **stream);
  void CreateStream(const WCHAR *name, IStream **stream);

  void CloseLastStorage();
  void CloseAllStorages() { CloseAllUpToSpecified(0); }
  void CloseAllExceptRoot() { CloseAllUpToSpecified(1); }
  void CloseAllUpToSpecified(int index);

  // void GetStorageStat(int i, CItemStat &stat) const;
  // CComBSTR GetStorageName(int i) const;

  void OpenStorageAlways(const WCHAR *name);
  void DestroyElement(const WCHAR *name);
};

HRESULT OpenRootStorage(const WCHAR *name, IStorage **newStorage, bool readOnly);
UInt64 GetStreamSize(IStream *stream);

#endif
