// CoderUtils.h

#ifndef __WINDOWS_RESOURCE_CODERUTILS_H
#define __WINDOWS_RESOURCE_CODERUTILS_H

#include "Common/MyString.h"
#include "Common/Types.h"

class CIndentMaker
{
  int m_NumIndents;
public:
  CIndentMaker(): m_NumIndents(0) {};
  void AddIndent() { m_NumIndents++;}
  void DeleteIndent() { m_NumIndents--;}
  void WriteIndent(UString &s)
  {
    for (int i = 0; i < m_NumIndents; i++)
      s += L"  ";
  }
};

class CUINT32FromToUStringMap
{
  // StandardResourceTypePairList m_Pairs;
  std::map<UINT32, UString> m_NumberToStringMap;
  std::map<UString, UINT32> m_StringToNumberMap;
public:
  void AddType(UINT32 number, const UString &s);
  bool GetString(UINT32 number, UString &s) const;
  bool GetNumber(const UString &s, UINT32 &number) const;
};

struct COptionPair
{
  UINT32 Flag;
  UString Name;
};

class COptionsCommon
{
  CObjectVector<COptionPair> m_Pairs;
public:
  void AddType(UINT32 number, wchar_t *name);
  void GetOptionNames(UINT32 flag, UStringVector &names, UINT32 &notUsedFlags) const;
  bool GetFirstZeroFlagName(UString &name) const;
  bool GetOptionFlag(const UString  &name, UINT32 &flag) const;
};

class CDataInStream
{
  UInt16 *m_Data;
  UINT32 m_Size;
  UINT32 m_Pos;
public:
  CDataInStream(UInt16 *data, UINT32 size): m_Data(data), m_Size(size), m_Pos(0) {};
  UINT16 GetWORD();
  UINT16 PeekWORD(UINT32 offset) const { return m_Data[m_Pos + offset]; };
  UINT32 GetTheRestSizeInWORDs() const { return m_Size - m_Pos; };
  UINT32 GetWordPosition() const { return m_Pos; };
  UINT32 GetBytePosition() const { return m_Pos * 2; };

  UINT32 GetDWORD();
  UString GetString();
  void SkipWords(UINT32 numWords);
  void SkipBytes(UINT32 numBytes);
  void SkipToAlignDWORD();
  const UInt16 *GetData() const { return &m_Data[m_Pos]; }
};

struct CDataOutStream
{
  CRecordVector<UInt16> m_Data;
  
  void PutWORD(UInt16 value);
  void PutDWORD(UInt32 value);
  void PutString(const UString &s);
  void AlignToDWORD();
};

class CTextInStream
{
  UString m_Text;
  int m_Pos;
  int m_Len;
public:
  void Init(const UINT16 *data, UINT32 size);
  bool WaitCharNotStepOver(wchar_t c);
  bool WaitChar(wchar_t c);
  bool WaitComma();
  bool WaitCommaNotStepOver();
  bool WaitOrDelimiter();
  bool GetSimpleString(UString &s);
  bool GetString(bool &inQuotes, UString &s);
};

UString UnsignedToUString(UINT32 value);
UString UnsignedToHexUString(UINT32 value);
CSysString DWORDToString(UINT32 value);

#endif
