// IconCursor.h

#ifndef __WINDOWS_RESOURCE_ICONCURSOR_H
#define __WINDOWS_RESOURCE_ICONCURSOR_H

#include "ResCoder.h"

namespace NWindows {
namespace NResource {
namespace NIconCursor {

#pragma pack( push, PragmaHeader)
#pragma pack( push, 1)

struct CHeader
{
  UINT16 Reserved;  // Reserved; must be zero
  UINT16 ResType;   // Specifies the resource type. This member must be RES_ICON
                  // for icons and RES_CURSOR for cursors.
                  // RES_ICON = 1
                  // RES_CURSOR = 2

  UINT16 ResCount;  // Specifies the number of icon or cursor components in the resource group
};

// One or more RESDIR structures immediately follow the NEWHEADER structure in
// the .res file. The ResCount member specifies the number of RESDIR structures

struct CIconResDir
{
  BYTE Width; // Specifies the width of the icon, in pixels. Acceptable values are 16, 32, and 64.
  BYTE Height; // Specifies the height of the icon, in pixels. Acceptable values are 16, 32, and 64.
  BYTE ColorCount;  // Specifies the number of colors in the icon. Acceptable values are 2, 8, and 16.
  BYTE Reserved;  // Reserved; must be set to the same value as that of the reserved field in the icon file header.
};
 
struct CCursorDirInCurFile
{
  BYTE Width; // Specifies the width of the icon, in pixels. Acceptable values are 16, 32, and 64.
  BYTE Height; // Specifies the height of the icon, in pixels. Acceptable values are 16, 32, and 64.
  BYTE Reserved1;  // Specifies the number of colors in the icon. Acceptable values are 2, 8, and 16.
  BYTE Reserved2;  // Reserved; must be set to the same value as that of the reserved field in the icon file header.
};

struct CCursorDirInGroupCur
{
  UINT16 Width; // Specifies the width of the icon, in pixels.
  UINT16 Height; // Specifies the height of the icon, in pixels.
};

struct CCursorSpots
{
  UINT16 XSpot;
  UINT16 YSpot;
};

struct CResDirInFile
{
  union
  {
    CIconResDir Icon;
    CCursorDirInCurFile Cursor;
  };
  union
  {
    UINT16 Reserved1; // Planes;  // Specifies the number of color planes in the icon or cursor bitmap.
    UINT16 XSpot;
  };
  union
  {
    UINT16 Reserved2; // BitCount; // Specifies the number of bits per pixel in the icon or cursor bitmap.
    UINT16 YSpot;
  };
  UINT32   BytesInRes; // Specifies the size of the resource, in bytes.
  UINT32 ImageOffset;
};

struct CResDirInGroup
{
  union
  {
    CIconResDir Icon;
    CCursorDirInGroupCur Cursor;
  };
  UINT16 Planes;  // Specifies the number of color planes in the icon or cursor bitmap.
  UINT16 BitCount; // Specifies the number of bits per pixel in the icon or cursor bitmap.
  UINT32 BytesInRes; // Specifies the size of the resource, in bytes.
  UINT16 ID; // Specifies the icon or cursor with a unique ordinal identifier.
};

// If the RESDIR structure contains information about a cursor,
// the first two WORDS the resource compiler writes to the RT_CURSOR
// resource are the xHotSpot and yHotSpot members of the LOCALHEADER structure.
 
#pragma pack(pop)
#pragma pack(pop, PragmaHeader)

}}}

#endif
