// Far/Resource/Main.cpp
// Check Align for updating !!!

#include "StdAfx.h"

#include "Common/StringConvert.h"
#include "Common/WildCard.h"

#include "Windows/FileDir.h"
#include "Windows/FileFind.h"
#include "Windows/FileIO.h"
#include "Windows/FileName.h"

#include "../../../7zip/UI/Far/FarUtils.h"

#include "CoderUtils.h"
#include "Messages.h"

#include "ResBitmap.h"
#include "ResDialog.h"
#include "ResGroupIcon.h"
#include "ResIcon.h"
#include "ResMenu.h"
#include "ResString.h"
#include "ResUtils.h"
#include "ResVersion.h"

#include "UpdateManager.h"

using namespace NFar;
using namespace NWindows;
using namespace NResource;

int MyStringCompareNoCase(const char *s1, const char  *s2);

int MyStringCompareNoCase(const char *s1, const char *s2)
{
  return MyStringCompareNoCase(MultiByteToUnicodeString(s1), MultiByteToUnicodeString(s2));
}

// User Data for
// NLevel::kLang is index of Lang in (CItemInfoDatabase)m_ItemInfoDatabase::m_LangIDs
// NLevel::kType is index of Type in (CItemInfoDatabase)m_ItemInfoDatabase::m_Types,
// NLevel::kItems is index of SubResource in (CSubResourceVector)m_SubResources

static const char *kConfigResourceOptionsCaption = "Resource browser 2";
static const char *kCommandPrefix = "Resource";

static const int kDescriptionMaxSize = 256;
static const char *kHelpTopicExtrFromResource =  "ExtractResources";
static const char *kHelpTopicConfig =  "ResourceConfig";

// static const char *kExractPathHistoryName = "ResourceExtactDestPath";

static const char *kPliginNameForRegestry = "Resource";

static const TCHAR *kRegisrtryMainKeyName = TEXT("");

static const TCHAR *kRegisrtryValueNameEnabled = TEXT("Enabled");
static bool kPluginEnabledDefault = true;

static const TCHAR *kRegisrtryValueNameStringsInUnicode = TEXT("TextInUnicode");
static bool kStringsInUnicodeModeDefault = true;


HINSTANCE g_hInstance;
#ifndef _UNICODE
bool g_IsNT = false;
static bool IsItWindowsNT()
{
  OSVERSIONINFO versionInfo;
  versionInfo.dwOSVersionInfoSize = sizeof(versionInfo);
  if (!::GetVersionEx(&versionInfo))
    return false;
  return (versionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT);
}
#endif

BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID)
{
  if (dwReason == DLL_PROCESS_ATTACH)
  {
    g_hInstance = hInstance;
    #ifndef _UNICODE
    g_IsNT = IsItWindowsNT();
    #endif
  }
  return TRUE;
}

static void ShowErrorMessage(const AString &message)
{
  g_StartupInfo.ShowMessage(message);
}

static bool IsOSHigherThanNT4()
{
  OSVERSIONINFO osInfo;
  osInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  if (GetVersionEx(&osInfo) == FALSE)
  {
    ShowErrorMessage("GetVersionEx fails");
    return false;
  }
  return (osInfo.dwPlatformId == VER_PLATFORM_WIN32_NT && osInfo.dwMajorVersion >= 4);
}
                                         // information structure);
struct COptions
{
  bool StringsInUnicodeMode;
  bool Enabled;
};

static COptions g_Options;

EXTERN_C void WINAPI SetStartupInfo(struct PluginStartupInfo *info)
{
  MY_TRY_BEGIN;
  g_StartupInfo.Init(*info, kPliginNameForRegestry);
  g_Options.StringsInUnicodeMode = g_StartupInfo.QueryRegKeyValue(
      HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameStringsInUnicode, kStringsInUnicodeModeDefault);
  g_Options.Enabled = g_StartupInfo.QueryRegKeyValue(
      HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameEnabled, kPluginEnabledDefault);
  MY_TRY_END1("Set Startup Info");
}


namespace NLevel
{
  enum EEnum
  {
    kLang,
    kType,
    kItems
  };
}

struct CSubResource
{
  int RecordIndex;
  CSubResourceProperties SubResourceProperties;
};

typedef CObjectVector<CSubResource> CSubResourceVector;

typedef CRecordVector<PluginPanelItem> CPluginPanelItemVector;

struct CResourceInFileInfo
{
  CSysString FileName;
  DWORD Size;
  bool IsDir;
};

typedef CObjectVector<CResourceInFileInfo> CResourceInFileInfoVector;

static CDefaultDecoder g_DefaultDecoder;
static NString::CDecoder g_StringDecoder;
static NBitmap::CDecoder g_BitmapDecoder;
static NIcon::CDecoder g_IconDecoder;
static NCursor::CDecoder g_CursorDecoder;
static NGroupIcon::CDecoder g_GroupIconDecoder;
static NGroupCursor::CDecoder g_GroupCursorDecoder;
static NMenu::CDecoder g_MenuDecoder;
static NVersion::CDecoder g_VersionDecoder;
static NDialog::CDecoder g_DialogDecoder;


// static CDefaultResourcePacker g_DefaultResourcePacker;
static NString::CEncoder g_StringEncoder;
static NBitmap::CEncoder g_BitmapEncoder;
static NIcon::CEncoder g_IconEncoder;
static NCursor::CEncoder g_CursorEncoder;
static NMenu::CEncoder g_MenuEncoder;

template <class T>
class CUniqueRecordVector: public CRecordVector<T>
{
public:
  int Find(const T& item) const
  {
    for (int i = 0; i < _size; i++)
      if (item == (*this)[i])
        return i;
    return -1;
  }
  int AddUnique(const T& item)
  {
    int index = Find(item);
    if (index >= 0)
      return index;
    return Add(item);
  }
};

typedef CUniqueRecordVector<int> CUniqueIntVector;

class CPlugin
{
private:
  CStandardTypesNames m_StandardTypesNames;
  CEncoderManager m_EncoderManager;
  
  bool m_ResourcesAreLoaded;

  CItemInfoDatabase m_ItemInfoDatabase;
  CSubResourceVector m_SubResources;
  // CLibrary m_Library;

  // CPluginPanelItemVector m_PluginPanelItems;
  // bool m_PluginPanelItemsWereAllocated;
    
  NLevel::EEnum m_Level;
  int m_CurrentLangIndex;
  int m_CurrentTypeIndex;

  CSysString m_CurrentDir;

  CSysString m_PannelTitle;
  
  InfoPanelLine m_InfoLines[1];

  char m_FileNameBuffer[1024];
  char m_CurrentDirBuffer[1024];
  char m_PannelTitleBuffer[1024];


  CSysString GetTypeName(const CItemTypeID &type) const;
  void GetTypeIndexesForOneLang(int langIndex,
      CUniqueIntVector &typeIndices) const;

  NFileOperationReturnCode::EEnum GetOneFile(const CSysString &destPath, int index);
  NFileOperationReturnCode::EEnum GetTypeItems(const CSysString &destPath,
      int langIndex, int typeIndex);
  NFileOperationReturnCode::EEnum GetLangItems(const CSysString &destPath,
      int langIndex);

  
  NFileOperationReturnCode::EEnum PutItems(CUpdateManager &manager,
      LANGID langID,
      const CItemTypeID &resourceType,
      const CResourceInFileInfoVector &items);

  NFileOperationReturnCode::EEnum PutTypeItem(CUpdateManager &manager,
      LANGID langID, const CSysString &typeDirectoryName);

  NFileOperationReturnCode::EEnum PutLangItems(CUpdateManager &manager, struct PluginPanelItem *panelItems,int numItems);
  NFileOperationReturnCode::EEnum PutTypeItems(CUpdateManager &manager, struct PluginPanelItem *panelItems, int numItems);
  NFileOperationReturnCode::EEnum PutPanelItems(CUpdateManager &manager, struct PluginPanelItem *panelItems, int numItems);

  bool DeleteOneRecord(CUpdateManager &manager, const CItemInfo &itemInfo);
  bool DeleteTypeItems(CUpdateManager &manager, int langIndex, int typeIndex);
  bool DeleteLangItems(CUpdateManager &manager, int langIndex);
  bool DeleteSubResurceItems(CUpdateManager &manager, CRecordVector<int> &subResourceIndexes);

  NFileOperationReturnCode::EEnum PutStringItem(const CSysString &fileName, DWORD fileSize,
      CSubResourceProperties &subResourceProps);

  NFileOperationReturnCode::EEnum PutStringItems(CUpdateManager &manager, LANGID langID,
      const CItemTypeID &resourceType, const CResourceInFileInfoVector &items);

public:
  CSysString m_FileName;

  CPlugin():m_Level(NLevel::kLang), m_ResourcesAreLoaded(false)
  {
    m_StandardTypesNames.AddEnglishNames();

    m_EncoderManager.AddEncoder(&g_StringEncoder);
    m_EncoderManager.AddEncoder(&g_BitmapEncoder);
    m_EncoderManager.AddEncoder(&g_IconEncoder);
    m_EncoderManager.AddEncoder(&g_CursorEncoder);
    m_EncoderManager.AddEncoder(&g_MenuEncoder);
    // m_EncoderManager.AddEncoder(&g_DefaultResourcePacker);
  };
  ~CPlugin(){};

  /*
  bool IsProcessed(const CSysString &fileName)
  {
    m_FileName = fileName;
    CModule m_Library;
    return m_Library.LoadEx(m_FileName);
  }
  */
  bool LoadAllResources();
  int GetFindData(PluginPanelItem **pPanelItem,int *pItemsNumber,int OpMode);
  void FreeFindData(PluginPanelItem *PanelItem,int ItemsNumber);
  int SetDirectory(char *dirPath, int opMode);
  void GetOpenPluginInfo(struct OpenPluginInfo *info);
  int DeleteFiles(PluginPanelItem *panelItems, int numItems, int opMode);
  int ProcessHostFile(struct PluginPanelItem *PanelItem,int ItemsNumber,int OpMode);
  NFileOperationReturnCode::EEnum GetFiles(struct PluginPanelItem *pi, int numItems,
      int moveMode, char *destPath, int opMode);

  NFileOperationReturnCode::EEnum PutFiles(struct PluginPanelItem *panelItems, int numItems,
      int moveMode, int opMode);
  int ProcessKey(int Key,unsigned int ControlState);
  static int SelectFormat(char *ArcFormat);
  static int FormatToPlugin(char *Format,int &PluginNumber,int &PluginType);
  void NormalizeState();
};

static HANDLE MyOpenFilePlugin(const char *name)
{
  CSysString normalizedName = GetSystemString(name); // check  OEM encodeng !!!
  normalizedName.Trim();
  CSysString fullName;
  NFile::NDirectory::MyGetFullPathName(normalizedName, fullName);
  {
    CModule m_Library;
    if (!m_Library.LoadEx(fullName))
      return (INVALID_HANDLE_VALUE);
  }
  CPlugin *plugin = new CPlugin();
  if (plugin == NULL)
    return(INVALID_HANDLE_VALUE);
  plugin->m_FileName = fullName;
  return (HANDLE)(plugin);
}

EXTERN_C HANDLE WINAPI OpenFilePlugin(char *name, const unsigned char * /* data */, unsigned int /* dataSize */)
{
  MY_TRY_BEGIN;

  if (name == NULL || (!g_Options.Enabled))
  {
    // if (!Opt.ProcessShiftF1)
      return(INVALID_HANDLE_VALUE);
  }
  return MyOpenFilePlugin(name);
  MY_TRY_END2("OpenFilePlugin", INVALID_HANDLE_VALUE);
}

EXTERN_C HANDLE WINAPI OpenPlugin(int openFrom, int item)
{
  MY_TRY_BEGIN;
  if (openFrom == OPEN_COMMANDLINE)
  {
    CSysString fileName = GetSystemString((const char *)item);
    if (fileName.IsEmpty())
      return INVALID_HANDLE_VALUE;
    if (fileName.Length() >= 2 &&
        fileName[0] == '\"' && fileName[fileName.Length() - 1] == '\"')
      fileName = fileName.Mid(1, fileName.Length() - 2);

    return MyOpenFilePlugin(fileName);
  }
  if (openFrom == OPEN_PLUGINSMENU)
  {
    PluginPanelItem pluginPanelItem;
    if (!g_StartupInfo.ControlGetActivePanelCurrentItemInfo(pluginPanelItem))
      throw 142134;
    return MyOpenFilePlugin(pluginPanelItem.FindData.cFileName);
  }
  return INVALID_HANDLE_VALUE;
  MY_TRY_END2("OpenPlugin", INVALID_HANDLE_VALUE);
}


EXTERN_C void WINAPI ClosePlugin(HANDLE hPlugin)
{
  MY_TRY_BEGIN;
  delete (CPlugin *)hPlugin;
  MY_TRY_END1("ClosePlugin");
}

EXTERN_C int WINAPI GetFindData(HANDLE _aPlugin, struct PluginPanelItem **panelItems, int *numItems, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)_aPlugin;
  return(plugin->GetFindData(panelItems, numItems, opMode));
  MY_TRY_END2("GetFindData", FALSE);
}

bool CPlugin::LoadAllResources()
{
  CModule m_Library;
  if (!m_Library.LoadEx(m_FileName))
    return false;

  CDecoderManager decoderManager;
  decoderManager.AddDecoder(&g_StringDecoder);
  decoderManager.AddDecoder(&g_BitmapDecoder);
  decoderManager.AddDecoder(&g_IconDecoder);
  decoderManager.AddDecoder(&g_CursorDecoder);
  decoderManager.AddDecoder(&g_GroupIconDecoder);
  decoderManager.AddDecoder(&g_GroupCursorDecoder);
  decoderManager.AddDecoder(&g_MenuDecoder);
  decoderManager.AddDecoder(&g_DialogDecoder);
  decoderManager.AddDecoder(&g_VersionDecoder);
  decoderManager.AddDecoder(&g_DefaultDecoder);

  m_SubResources.Clear();
  m_Library.ReadItemInfoDatabase(m_ItemInfoDatabase);
  for (int i = 0; i < m_ItemInfoDatabase.Items.Size(); i++)
  {
    CDataInfo resourceData;
    if (!m_Library.ReadResource(resourceData, m_ItemInfoDatabase, i))
      continue; // test it
    CSubResourcePropertiesVector subResourcePropertiesVector;
    decoderManager.ReadItems(m_ItemInfoDatabase, i, resourceData, subResourcePropertiesVector);
    for (int j = 0; j < subResourcePropertiesVector.Size(); j++)
    {
      m_SubResources.Add(CSubResource());
      CSubResource &subResource = m_SubResources.Back();
      subResource.RecordIndex = i;
      subResource.SubResourceProperties = subResourcePropertiesVector[j];
    }
  }
  m_ResourcesAreLoaded = true;
  return true;
}

static CSysString GetLangName(LANGID langID)
{
  return DWORDToString(langID);
}

bool MyGetLocaleInfo(LCID aLCID, LCTYPE aLCTYPE, CSysString &result)
{
  // LCTYPE

  // LOCALE_IDEFAULTLANGUAGE
  // Language identifier for the principal language spoken in
  // this locale. This is provided so partially specified locales
  // can be completed with default values. The maximum number of
  // characters allowed for this string is five.

  // LOCALE_ILANGUAGE
  // Language identifier indicating the language.
  // The maximum number of characters allowed for this string is five.

  // LOCALE_SABBREVLANGNAME
  // Abbreviated name of the language, created by taking the two-letter
  // language abbreviation from the ISO Standard 639 and adding a third letter,
  // as appropriate, to indicate the sublanguage.

  // LOCALE_SENGLANGUAGE
  // Full English name of the language from the International Organization
  // for Standardization (ISO) Standard 639. This is always restricted to
  // characters that can be mapped into the ASCII 127-character subset.

  // LOCALE_SISO639LANGNAME
  // Windows NT 4.0 or later: The abbreviated name of the
  // language based on the ISO Standard 639 values.

  
  // LOCALE_SLANGUAGE
  // Full localized name of the language.

  // LOCALE_SNATIVELANGNAME
  // Native name of the language.

  
  int size = ::GetLocaleInfo(aLCID, aLCTYPE, NULL, 0);
  if (size == 0)
    return false;
  size = ::GetLocaleInfo(aLCID, aLCTYPE, result.GetBuffer(size), size);
  result.ReleaseBuffer();
  return (size != 0);
}

static CSysString GetLangDescription(LANGID langID)
{
  LCTYPE aLCTYPE = LOCALE_SLANGUAGE;
  LCID aLCID = MAKELANGID(langID, SORT_DEFAULT);

  CSysString s;
  if (!MyGetLocaleInfo(aLCID, aLCTYPE, s))
    s = "Unknown";
  return s;
}

CSysString CPlugin::GetTypeName(const CItemTypeID &type) const
{
  CSysString name;
  if (type.IsNumber() && m_StandardTypesNames.GetTypeName(type.GetAsNumber(), name))
    return name;
  return type.GetAsString();
}

void CPlugin::GetTypeIndexesForOneLang(int langIndex, CUniqueIntVector &typeIndices) const
{
  typeIndices.Clear();
  for (int i = 0; i < m_ItemInfoDatabase.Items.Size(); i++)
  {
    const CItemInfo &itemInfo = m_ItemInfoDatabase.Items[i];
    if (itemInfo.LangIndex == langIndex)
      typeIndices.AddUnique(itemInfo.TypeIndex);
  }
}

static bool IsTextUnicodeType(const CItemTypeID &type)
{
  return (
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_STRING) ||
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_MENU) ||
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_GROUP_ICON) ||
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_GROUP_CURSOR) ||
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_DIALOG) ||
      (type.IsNumber() && type.GetAsLPCTSTR() == RT_VERSION)
    );
}

static bool InUnicode(const CItemTypeID &type)
{
  return (g_Options.StringsInUnicodeMode && IsTextUnicodeType(type));
}


static bool InNonUnicode(const CItemTypeID &type)
{
  return (!g_Options.StringsInUnicodeMode && IsTextUnicodeType(type));
}

CSysString ConvertAnsiToOem(const CSysString &ansiString)
{
  CSysString oemString;
  CharToOem(ansiString, oemString.GetBuffer(ansiString.Length() * 2 + 1));
  oemString.ReleaseBuffer();
  return oemString;
}
 
int CPlugin::GetFindData(PluginPanelItem **panelItems, int *numItemsRes, int opMode)
{
  /*
  if (m_PluginPanelItemsWereAllocated)
  {
    ShowErrorMessage("GetFindData error");
    return FALSE;
  }
  */

  NFile::NFind::CFileInfo fi;
  if (!fi.Find(m_FileName)) // test it
    return FALSE;

  /*
  *panelItems = NULL;
  *numItemsRes = 0;

  PluginPanelItem pi;
  if (m_CurrentDir.IsEmpty())
    pi.FindData.dwFileAttributes = FILE_ATTRIBUTE_DIRECTORY;
  else
    pi.FindData.dwFileAttributes = 0;
  strcpy(pi.FindData.cFileName, "yyy");
  strcpy(pi.FindData.cAlternateFileName,
    pi.FindData.cFileName); // change it
  pi.FindData.ftCreationTime = *((FILETIME*) (&fi.CTime));
  pi.FindData.ftLastAccessTime = *((FILETIME*) (&fi.ATime));
  pi.FindData.ftLastWriteTime = *((FILETIME*) (&fi.MTime));
  pi.FindData.nFileSizeHigh = 0;
  pi.FindData.nFileSizeLow = 0;
  pi.PackSizeHigh = pi.FindData.nFileSizeHigh;
  pi.PackSize = pi.FindData.nFileSizeLow;
  pi.Flags = 0;
  pi.NumberOfLinks = 0;
  pi.Description = NULL;
  pi.Owner = NULL;
  pi.CustomColumnData = NULL;
  pi.CustomColumnNumber = 0;
  pi.UserData = 0;
  {
    int numItems = 1;
    *panelItems = new PluginPanelItem[numItems];
    memcpy(*panelItems, &pi, sizeof(PluginPanelItem) * numItems);
    *numItemsRes = numItems;
  }
  return TRUE;
  */

  CScreenRestorer screenRestorer;
  if (!m_ResourcesAreLoaded)
  {
    if ((opMode & OPM_SILENT) == 0 && (opMode & OPM_FIND ) == 0)
    {
      screenRestorer.Save();
      const char *msgItems[]=
      {
        g_StartupInfo.GetMsgString(kFarResMsgArcReadTitle),
        g_StartupInfo.GetMsgString(kFarResMsgArcReading)
      };
      g_StartupInfo.ShowMessage(0, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 0);
    }
    LoadAllResources();
  }


  CPluginPanelItemVector pluginPanelItems;
  
  // m_PluginPanelItems.Clear();
  switch(m_Level)
  {
    case NLevel::kLang:
    {
      for (int i = 0; i < m_ItemInfoDatabase.LangIDs.Size(); i++)
      {
        PluginPanelItem pi;
        pi.FindData.dwFileAttributes = FILE_ATTRIBUTE_DIRECTORY;
        strcpy(pi.FindData.cFileName, GetLangName(m_ItemInfoDatabase.LangIDs[i]));
        pi.FindData.cAlternateFileName[0] = 0;

        pi.FindData.ftCreationTime = *((FILETIME*) (&fi.CTime));
        pi.FindData.ftLastAccessTime = *((FILETIME*) (&fi.ATime));
        pi.FindData.ftLastWriteTime = *((FILETIME*) (&fi.MTime));
        pi.FindData.nFileSizeHigh = 0;
        pi.FindData.nFileSizeLow = 0;
        pi.PackSizeHigh = pi.FindData.nFileSizeHigh;
        pi.PackSize = pi.FindData.nFileSizeLow;
        pi.Flags = 0;
        pi.NumberOfLinks = 0;
        CSysString description = GetLangDescription(m_ItemInfoDatabase.LangIDs[i]);
        CSysString oemDescription = ConvertAnsiToOem(description.Left(kDescriptionMaxSize));
        pi.Description = new char[oemDescription.Length() + 1];
        strcpy(pi.Description, oemDescription);
        pi.Owner = NULL;
        pi.CustomColumnData = NULL;
        pi.CustomColumnNumber = 0;
        pi.UserData = i;
        pluginPanelItems.Add(pi);
      }
      break;
    }
    case NLevel::kType:
    {
      CUniqueIntVector typeIndices;
      GetTypeIndexesForOneLang(m_CurrentLangIndex, typeIndices);
      for (int i = 0; i < typeIndices.Size(); i++)
      {
        int typeIndex = typeIndices[i];
        PluginPanelItem pi;
        pi.FindData.dwFileAttributes = FILE_ATTRIBUTE_DIRECTORY;
        
        CSysString name = GetTypeName(m_ItemInfoDatabase.TypeIDs[typeIndex]);

        strcpy(pi.FindData.cFileName, name);
        pi.FindData.cAlternateFileName[0] = 0;
        pi.FindData.ftCreationTime = *((FILETIME*) (&fi.CTime));
        pi.FindData.ftLastAccessTime = *((FILETIME*) (&fi.ATime));
        pi.FindData.ftLastWriteTime = *((FILETIME*) (&fi.MTime));
        pi.FindData.nFileSizeHigh = 0;
        pi.FindData.nFileSizeLow = 0;
        pi.PackSizeHigh = pi.FindData.nFileSizeHigh;
        pi.PackSize = pi.FindData.nFileSizeLow;
        pi.Flags = 0;
        pi.NumberOfLinks = 0;
        pi.Description = NULL;
        pi.Owner = NULL;
        pi.CustomColumnData = NULL;
        pi.CustomColumnNumber = 0;
        pi.UserData = typeIndex;
        pluginPanelItems.Add(pi);
      }
      break;
    }
    case NLevel::kItems:
    {
      for (int i = 0; i < m_SubResources.Size(); i++)
      {
        const CSubResource &subResource = m_SubResources[i];

        const CItemInfo &itemInfo =
            m_ItemInfoDatabase.Items[subResource.RecordIndex];
        if (itemInfo.LangIndex != m_CurrentLangIndex ||
            itemInfo.TypeIndex != m_CurrentTypeIndex)
          continue;

        PluginPanelItem pi;
        pi.FindData.dwFileAttributes = 0; // change it;
        
        const CSubResourceProperties &subResourceProps = subResource.SubResourceProperties;

        CSysString name = subResourceProps.GetFullName();

        strcpy(pi.FindData.cFileName, name);
        pi.FindData.cAlternateFileName[0] = 0;

        DWORD resourceSize = subResourceProps.Size;
        const CItemTypeID &type = m_ItemInfoDatabase.TypeIDs[itemInfo.TypeIndex];

        if (InNonUnicode(type) && !subResourceProps.DefaultExtension.IsEmpty())
           resourceSize /= sizeof(wchar_t);

        pi.FindData.ftCreationTime = *((FILETIME*) (&fi.CTime));
        pi.FindData.ftLastAccessTime = *((FILETIME*) (&fi.ATime));
        pi.FindData.ftLastWriteTime = *((FILETIME*) (&fi.MTime));
        pi.FindData.nFileSizeHigh = 0;
        pi.FindData.nFileSizeLow = resourceSize;
        pi.PackSizeHigh = pi.FindData.nFileSizeHigh;
        pi.PackSize = pi.FindData.nFileSizeLow;
        pi.Flags = 0;
        pi.NumberOfLinks = 0;

        const CSysString &description = subResourceProps.Description;
        if (description.IsEmpty())
          pi.Description = NULL;
        else
        {
          CSysString oemDescription = ConvertAnsiToOem(description.Left(kDescriptionMaxSize));
          pi.Description = new char[oemDescription.Length() + 1];
          strcpy(pi.Description, oemDescription);
        }
        pi.Owner = NULL;
        pi.CustomColumnData = NULL;
        pi.CustomColumnNumber = 0;
        pi.UserData = i;
        pluginPanelItems.Add(pi);
      }
      break;
    }
  }
  int numItems = pluginPanelItems.Size();
  *panelItems = new PluginPanelItem[numItems];
  memcpy(*panelItems, &pluginPanelItems.Front(), sizeof(PluginPanelItem) * numItems);
  *numItemsRes = numItems;
  // m_PluginPanelItemsWereAllocated = true;

  return TRUE;
}

EXTERN_C void WINAPI FreeFindData(HANDLE _aPlugin, struct PluginPanelItem *panelItems, int numItems)
{
  // MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)_aPlugin;
  plugin->FreeFindData(panelItems, numItems);
  // MY_TRY_END1("FreeFindData");
}

void CPlugin::FreeFindData(struct PluginPanelItem *panelItems, int numItems)
{
  for (int i = 0; i < numItems; i++)
    if (panelItems[i].Description != NULL)
      delete []panelItems[i].Description;
  delete []panelItems;
}


EXTERN_C int WINAPI GetFiles(HANDLE hPlugin, struct PluginPanelItem *panelItems,
    int numItems, int moveMode, char *destPath, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->GetFiles(panelItems, numItems, moveMode, destPath, opMode));
  MY_TRY_END2("GetFiles", NFileOperationReturnCode::kError);
}

NFileOperationReturnCode::EEnum CPlugin::GetOneFile(const CSysString &destPath, int index)
{
  const CSubResource &subResource = m_SubResources[index];
  
  const CItemInfo &itemInfo = m_ItemInfoDatabase.Items[subResource.RecordIndex];
  const CSubResourceProperties &subResourceProps = subResource.SubResourceProperties;
  CSysString name = subResourceProps.GetFullName();
  
  DWORD resourceSize = subResourceProps.Size;
  const CItemTypeID &type = m_ItemInfoDatabase.TypeIDs[itemInfo.TypeIndex];
  const void *ptr = subResourceProps.Data;

  CSysString charDynaicBuffer;
  UINT32 outSize = resourceSize;
  if (InNonUnicode(type) && !subResourceProps.DefaultExtension.IsEmpty())
  {
    resourceSize /= sizeof(wchar_t);
    if (resourceSize != 0)
    {
      int numRequiredBytes = resourceSize * 2;
      outSize = ::WideCharToMultiByte(CP_ACP, 0,
          (const wchar_t *)((const BYTE *)subResourceProps.Data),
          resourceSize, charDynaicBuffer.GetBuffer(numRequiredBytes),
          numRequiredBytes + 1, NULL, NULL);
      charDynaicBuffer.ReleaseBuffer(outSize);
      if (resourceSize != outSize)
      {
        ShowErrorMessage("Internal error #1518");
        return NFileOperationReturnCode::kError;
      }
    }
    ptr = (LPCTSTR)charDynaicBuffer;
  }

  NFile::NIO::COutFile inFile;
  if (!inFile.Create(destPath + name, true))
  {
    ShowErrorMessage("Create error");
    return NFileOperationReturnCode::kError;
  }
  UINT32 processedSize;
  if (!inFile.Write(ptr, outSize, processedSize) ||
      processedSize != outSize)
  {
    ShowErrorMessage("Write error");
    return NFileOperationReturnCode::kError;
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::GetTypeItems(const CSysString &destPath,
    int langIndex, int typeIndex)
{
  const CItemTypeID &type = m_ItemInfoDatabase.TypeIDs[typeIndex];
  CSysString newPath = destPath + GetTypeName(type);
  NFile::NDirectory::CreateComplexDirectory(newPath); // can be optimized
  newPath += NFile::NName::kDirDelimiter;
  for (int i = 0; i < m_SubResources.Size(); i++)
  {
    const CSubResource &subResource = m_SubResources[i];
    const CItemInfo &itemInfo = m_ItemInfoDatabase.Items[subResource.RecordIndex];
    if (itemInfo.LangIndex != langIndex || itemInfo.TypeIndex != typeIndex)
      continue;
    NFileOperationReturnCode::EEnum result = GetOneFile(newPath, i);
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::GetLangItems(const CSysString &destPath, int langIndex)
{
  CSysString newPath = destPath + GetLangName(m_ItemInfoDatabase.LangIDs[langIndex]);
  NFile::NDirectory::CreateComplexDirectory(newPath); // can be optimized
  newPath += NFile::NName::kDirDelimiter;
  CUniqueIntVector typeIndices;
  GetTypeIndexesForOneLang(langIndex, typeIndices);
  for (int i = 0; i < typeIndices.Size(); i++)
  {
    NFileOperationReturnCode::EEnum result = GetTypeItems(newPath, langIndex, typeIndices[i]);
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::GetFiles(struct PluginPanelItem *panelItems,
    int numItems, int moveMode, char *_aDestPath, int opMode)
{
  if (!IsOSHigherThanNT4() && moveMode != 0)
  {
    g_StartupInfo.ShowMessage(kFarResUpdatingOnlyUnderNT);
    return NFileOperationReturnCode::kError;
  }

  CScreenRestorer screenRestorer;
  if ((opMode & OPM_SILENT) == 0 && (opMode & OPM_FIND ) == 0)
  {
    screenRestorer.Save();
    const char *msgItems[]=
    {
      g_StartupInfo.GetMsgString(kFarResMsgArcReadTitle),
      g_StartupInfo.GetMsgString(kFarResMsgArcGetting)
    };
    g_StartupInfo.ShowMessage(0, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 0);
  }

  // CSysString destPath = _aDestPath;
  //NormalizeFilePrefix(destPath);
  CSysString destPath = _aDestPath;
  NFile::NName::NormalizeDirPathPrefix(destPath);

  if ((opMode & OPM_SILENT) == 0)
  {
    const int kPathIndex = 2;
    const int kOkButtonIndex = 4;
    for (;;)
    {
      struct CInitDialogItem initItems[]={
        { DI_DOUBLEBOX, 3, 1, 72, 6, false, false, 0, false, kFarResMsgExtractTitle, NULL, NULL },
        { DI_TEXT, 5, 2, 0, 0, false, false, 0, false, kFarResMsgExtractTo, NULL, NULL },
        
        //{ DI_EDIT, 5, 3, 70, 3, true, false, DIF_HISTORY, false, -1, destPath, kExractPathHistoryName},
        { DI_EDIT, 5, 3, 70, 3, true, false, 0, false, -1, LPCTSTR(destPath), NULL},
        
        { DI_TEXT, 3, 4, 0, 0, false, false, DIF_BOXCOLOR|DIF_SEPARATOR, false, -1, "", NULL  },
        { DI_BUTTON, 0, 5, 0, 0, false, false, DIF_CENTERGROUP, true, kFarResMsgExtrExtract, NULL, NULL  },
        { DI_BUTTON, 0, 5, 0, 0, false, false, DIF_CENTERGROUP, false, kFarResMsgExtrCancel, NULL, NULL  }
      };
      
      const int kNumDialogItems = sizeof(initItems) / sizeof(initItems[0]);
      FarDialogItem dialogItems[kNumDialogItems];
      g_StartupInfo.InitDialogItems(initItems, dialogItems, kNumDialogItems);
      
      int askCode = g_StartupInfo.ShowDialog(76, 8,
        kHelpTopicExtrFromResource, dialogItems, kNumDialogItems);
      if (askCode != kOkButtonIndex)
        return NFileOperationReturnCode::kInterruptedByUser;
      destPath = dialogItems[kPathIndex].Data;
      destPath.Trim();
      if (!destPath.IsEmpty() && destPath[destPath.Length() - 1] == NFile::NName::kDirDelimiter)
        break;
      g_StartupInfo.ShowMessage("You must specify directory path");
    }
  }

  NFile::NDirectory::CreateComplexDirectory(destPath);
  CSysString destPath2 = destPath;

  for (int i = 0; i < numItems; i++)
  {
    int index = panelItems[i].UserData;
    NFileOperationReturnCode::EEnum result;
    switch(m_Level)
    {
      case NLevel::kLang: result = GetLangItems(destPath2, index); break;
      case NLevel::kType: result = GetTypeItems(destPath2, m_CurrentLangIndex, index); break;
      case NLevel::kItems: result = GetOneFile(destPath2, index); break;
      default: return NFileOperationReturnCode::kError;
    }
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  if (moveMode != 0)
  {
    if (DeleteFiles(panelItems, numItems, opMode) == FALSE)
      return NFileOperationReturnCode::kError;
  }
  return NFileOperationReturnCode::kSuccess;
}

EXTERN_C int WINAPI SetDirectory(HANDLE hPlugin, char *dirPath, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->SetDirectory(dirPath, opMode));
  MY_TRY_END2("SetDirectory", FALSE);
}

static const wchar_t kDirDelimiter1 = L'\\';
static const wchar_t kDirDelimiter2 = L'/';

static inline bool IsCharDirLimiter(wchar_t c)
{
  return (c == kDirDelimiter1 || c == kDirDelimiter2);
}

void SplitPathToParts(const AString &path, AStringVector &pathParts)
{
  pathParts.Clear();
  AString name;
  int len = path.Length();
  if (len == 0)
    return;
  for (int i = 0; i < len; i++)
  {
    wchar_t c = path[i];
    if (IsCharDirLimiter(c))
    {
      pathParts.Add(name);
      name.Empty();
    }
    else
      name += path[i];
  }
  pathParts.Add(name);
}

int CPlugin::SetDirectory(char *aszDir, int /* opMode */)
{
  CSysString dirPath = aszDir;
  if (dirPath == "\\")
  {
    m_CurrentDir.Empty();
    m_Level = NLevel::kLang;
    return TRUE;
  }
  if (dirPath == "..")
  {
    if (m_CurrentDir.IsEmpty())
    {
      ShowErrorMessage("#1");
      return FALSE;
    }
    int delimiterIndex = m_CurrentDir.ReverseFind('\\');
    if (delimiterIndex >= 0)
      m_CurrentDir = m_CurrentDir.Left(delimiterIndex);
    else
      m_CurrentDir.Empty();
    switch(m_Level)
    {
      case NLevel::kLang:
      {
        ShowErrorMessage("#2");
        return FALSE;
      }
      case NLevel::kType:
      {
        m_Level = NLevel::kLang;
        return TRUE;
      }
      case NLevel::kItems:
      {
        m_Level = NLevel::kType;
        return TRUE;
      }
    }
  }
  else
  {
    CSysStringVector pathParts;
    SplitPathToParts(dirPath, pathParts);
    for (int i = 0; i < pathParts.Size(); i++)
    {
      CSysString name = pathParts[i];
      switch(m_Level)
      {
        case NLevel::kLang:
        {
          int j;
          for (j = 0; j < m_ItemInfoDatabase.LangIDs.Size(); j++)
          {
            CSysString langName = GetLangName(m_ItemInfoDatabase.LangIDs[j]);
            if (langName.CompareNoCase(name) == 0)
            {
              m_Level = NLevel::kType;
              m_CurrentLangIndex = j;
              if (!m_CurrentDir.IsEmpty())
                m_CurrentDir += NFile::NName::kDirDelimiter;
              m_CurrentDir += langName;
              break;
            }
          }
          if (j == m_ItemInfoDatabase.LangIDs.Size())
          {
            ShowErrorMessage("#3");
            return FALSE;
          }
          break;
        }
        case NLevel::kType:
        {
          int j;
          for (j = 0; j < m_ItemInfoDatabase.TypeIDs.Size(); j++)
          {
            CSysString typeName = GetTypeName(m_ItemInfoDatabase.TypeIDs[j]);
            if (typeName.CompareNoCase(name) == 0)
            {
              m_Level = NLevel::kItems;
              m_CurrentTypeIndex = j;
              m_CurrentDir += NFile::NName::kDirDelimiter;
              m_CurrentDir += typeName;
              break;
            }
          }
          if (j == m_ItemInfoDatabase.TypeIDs.Size())
          {
            ShowErrorMessage("#4");
            return FALSE;
          }
          break;
        }
        default:
        {
          ShowErrorMessage("Can not change directory");
          return FALSE;
        }
      }
    }
  }
  return TRUE;
}



EXTERN_C void WINAPI GetPluginInfo(struct PluginInfo *info)
{
  MY_TRY_BEGIN;
  info->StructSize = sizeof(*info);
  info->Flags = 0;
  info->DiskMenuStrings = NULL;
  info->DiskMenuNumbers = NULL;
  info->DiskMenuStringsNumber = 0;

  static const char *pluginMenuStrings[1];
  pluginMenuStrings[0] = g_StartupInfo.GetMsgString(kFarResMsgCfgLine);
  info->PluginMenuStrings = (char **)pluginMenuStrings;
  info->PluginMenuStringsNumber = 1;

  static const char *PluginCfgStrings[1];
  PluginCfgStrings[0] = g_StartupInfo.GetMsgString(kFarResMsgCfgLine);
  info->PluginConfigStrings = (char **)PluginCfgStrings;
  info->PluginConfigStringsNumber = sizeof(PluginCfgStrings) / sizeof(PluginCfgStrings[0]);
  info->CommandPrefix = (char *)kCommandPrefix;
  MY_TRY_END1("GetPluginInfo");
}

/*
enum EOptionsOutputHide
{
  kOptionsOutputHideNone,
  kOptionsOutputHideView,
  kOptionsOutputHideAlways
};
*/

EXTERN_C int WINAPI Configure(int /* itemNumber */)
{
  MY_TRY_BEGIN;

  const int kEnabledCheckBoxIndex = 1;
  const int kUnicodeCheckBoxIndex = 3;

  const int kYSize = 9;

  struct CInitDialogItem initItems[]=
  {
    { DI_DOUBLEBOX, 3, 1, 72, kYSize - 2, false, false, 0, false, kFarResMsgConfigTitle, NULL, NULL },
    // { DI_RADIOBUTTON, 5, 2, 0, 0, true , false, DIF_GROUP, false, kFarResMsgConfigHideNone, NULL, NULL },
    // { DI_RADIOBUTTON, 5, 3, 0, 0, false, false, 0, false, kFarResMsgConfigHideView, NULL, NULL },
    // { DI_RADIOBUTTON, 5, 4, 0, 0, false, false, 0, false, kFarResMsgConfigHideAlways, NULL, NULL },
    { DI_CHECKBOX, 5, 2, 0, 0, true, g_Options.Enabled, 0, false, kFarResMsgConfigPluginEnabled, NULL, NULL },
    { DI_TEXT, 5, 3, 0, 0, false, false, DIF_BOXCOLOR | DIF_SEPARATOR, false, -1, "", NULL },
    { DI_CHECKBOX, 5, 4, 0, 0, false, g_Options.StringsInUnicodeMode, 0, false, kFarResMsgConfigStringsInUnicode, NULL, NULL },
    { DI_TEXT, 5, 5, 0, 0, false, false, DIF_BOXCOLOR | DIF_SEPARATOR, false, -1, "", NULL },
    { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, true, kFarResMsgOk, NULL, NULL },
    { DI_BUTTON, 0, kYSize - 3, 0, 0, false, false, DIF_CENTERGROUP, false, kFarResMsgCancel, NULL, NULL },
  };

  const int kNumDialogItems = sizeof(initItems) / sizeof(initItems[0]);
  const int kOkButtonIndex = kNumDialogItems - 2;

  FarDialogItem dialogItems[kNumDialogItems];
  g_StartupInfo.InitDialogItems(initItems, dialogItems, kNumDialogItems);

  /*
  DialogItems[1].Selected=(Opt.HideOutput==0);
  DialogItems[2].Selected=(Opt.HideOutput==1);
  DialogItems[3].Selected=(Opt.HideOutput==2);
  DialogItems[5].Selected=Opt.ProcessShiftF1;
  strcpy(DialogItems[8].Data,Opt.DescriptionNames);
  DialogItems[10].Selected=Opt.ReadDescriptions;
  DialogItems[11].Selected=Opt.UpdateDescriptions;
  */

  int askCode = g_StartupInfo.ShowDialog(76, kYSize, kHelpTopicConfig, dialogItems, kNumDialogItems);

  if (askCode != kOkButtonIndex)
    return FALSE;

  g_Options.Enabled = BOOLToBool(dialogItems[kEnabledCheckBoxIndex].Selected);
  g_Options.StringsInUnicodeMode = BOOLToBool(dialogItems[kUnicodeCheckBoxIndex].Selected);

  g_StartupInfo.SetRegKeyValue(HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameEnabled, g_Options.Enabled);
  g_StartupInfo.SetRegKeyValue(HKEY_CURRENT_USER, kRegisrtryMainKeyName,
      kRegisrtryValueNameStringsInUnicode, g_Options.StringsInUnicodeMode);
  return TRUE;
  MY_TRY_END2("Configure", FALSE);
}

EXTERN_C void WINAPI GetOpenPluginInfo(HANDLE hPlugin,struct OpenPluginInfo *info)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  plugin->GetOpenPluginInfo(info);
  MY_TRY_END1("GetOpenPluginInfo");
}

static char *kPluginFormatName = "Resource";

void CPlugin::GetOpenPluginInfo(struct OpenPluginInfo *info)
{
  info->StructSize = sizeof(*info);
  info->Flags = OPIF_USEFILTER | OPIF_USESORTGROUPS| OPIF_USEHIGHLIGHTING | OPIF_ADDDOTS | OPIF_COMPAREFATTIME;

  strcpy(m_FileNameBuffer, m_FileName);
  info->HostFile = m_FileNameBuffer; // test it it is not static
  
  strcpy(m_CurrentDirBuffer, m_CurrentDir);
  info->CurDir = m_CurrentDirBuffer;

  info->Format = kPluginFormatName;

  //CSysString prefix, name;
  //SplitPathToPrefixAndName(m_FileName, prefix, name);

  CSysString name;
  {
    CSysString fullName;
    int index;
    NFile::NDirectory::MyGetFullPathName(m_FileName, fullName, index);
    name = fullName.Mid(index);
  }

  m_PannelTitle = "Resource:" + name;
  if (!m_CurrentDir.IsEmpty())
  {
    m_PannelTitle += '\\';
    m_PannelTitle += m_CurrentDir;
  }
 
  strcpy(m_PannelTitleBuffer, m_PannelTitle);
  info->PanelTitle = m_PannelTitleBuffer;

  strcpy(m_InfoLines[0].Text, "Resource Type File");
  strcpy(m_InfoLines[0].Data, "Normal DLL or EXE");
  m_InfoLines[0].Separator = 0;

  info->InfoLines = m_InfoLines;
  info->InfoLinesNumber = sizeof(m_InfoLines) / sizeof(m_InfoLines[0]);

  
  info->DescrFiles = NULL;
  info->DescrFilesNumber = 0;

  info->PanelModesArray = NULL;
  info->PanelModesNumber = 0;

  info->StartPanelMode = 0;
  info->StartSortMode = 0;
  info->KeyBar = NULL;
  info->ShortcutData = NULL;
}

EXTERN_C int WINAPI PutFiles(HANDLE hPlugin, struct PluginPanelItem *panelItems,
    int numItems, int moveMode, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin=(CPlugin *)hPlugin;
  return(plugin->PutFiles(panelItems, numItems, moveMode, opMode));
  MY_TRY_END2("PutFiles", NFileOperationReturnCode::kError);
}

WORD GetStringResourceLogicIDNumber(const CSysString &name)
{
  WORD number;
  if (!NResource::StringToUINT16(name, number))
    throw 251533;
  return number;
}

WORD GetStringResourceIDNumber(const CSysString &name)
{
  int dotIndex = name.ReverseFind('.');
  if (dotIndex < 0)
    throw 251530;
  CSysString pureName = name.Left(dotIndex);
  CSysString extension = name.Mid(dotIndex + 1);
  WORD number;
  if (!StringToUINT16(pureName, number))
    throw 251533;
  static WORD kBunchSize = 0x10;
  return (number / kBunchSize) + 1;
}

NFileOperationReturnCode::EEnum CPlugin::PutStringItem(const CSysString &fileName,
    DWORD fileSize, CSubResourceProperties &subResourceProps)
{
  CByteDynamicBuffer buffer;
  buffer.EnsureCapacity(fileSize);
  NFile::NIO::CInFile inFile;
  inFile.Open(fileName);
  UINT32 processedSize;
  if (!inFile.Read(buffer, fileSize, processedSize) || processedSize != fileSize)
  {
    g_StartupInfo.ShowMessage(kFarResFileReadingError);
    return NFileOperationReturnCode::kError;
  }

  CByteDynamicBuffer unicodeBuffer;

  if (g_Options.StringsInUnicodeMode && fileSize >= 2 && *((wchar_t *)((BYTE *)buffer)) == 0xFEFF)
  {
    subResourceProps.Size = fileSize - 2;
    unicodeBuffer.EnsureCapacity(subResourceProps.Size);
    memcpy(unicodeBuffer, buffer + 2, subResourceProps.Size);
    subResourceProps.Data = unicodeBuffer;
  }
  else if (!g_Options.StringsInUnicodeMode)
  {
    subResourceProps.Size = fileSize * sizeof(wchar_t);
    unicodeBuffer.EnsureCapacity(subResourceProps.Size);
    if (MultiByteToWideChar(CP_ACP, 0,
      (const char *)((const BYTE *)buffer), fileSize,
      (wchar_t *)((BYTE *)unicodeBuffer), fileSize) != int(fileSize))
    {
      ShowErrorMessage("Internal error #2337");
      return NFileOperationReturnCode::kError;
    }
    subResourceProps.Data = unicodeBuffer;
  }
  else
  {
    subResourceProps.Data = buffer;
    subResourceProps.Size = fileSize;
  }

  int dotIndex = fileName.ReverseFind('.');
  CSysString resourceNameString = fileName.Left(dotIndex);
  CSysString extension = fileName.Mid(dotIndex + 1);
  subResourceProps.DefaultExtension = extension;
  subResourceProps.Name.SetAsNumber(
      GetStringResourceLogicIDNumber(resourceNameString));
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::PutStringItems(CUpdateManager &manager,
    LANGID langID,
    const CItemTypeID &resourceType,
    const CResourceInFileInfoVector &items)
{
  CUniqueRecordVector<WORD> idNumberList;

  int i;
  for (i = 0; i < items.Size(); i++)
  {
    const CResourceInFileInfo &item = items[i];
    if (item.IsDir)
      throw 251714;
    const CSysString &fileName = item.FileName;
    WORD number = GetStringResourceIDNumber(fileName);
    idNumberList.AddUnique(number);
  }

  for (i = 0; i < idNumberList.Size(); i++)
  {
    WORD idNumber = idNumberList[i];
    CSubResourcePropertiesVector structure;

    int j;
    for (j = 0; j < items.Size(); j++)
    {
      const CResourceInFileInfo &item = items[j];
      const CSysString &fileName = item.FileName;
      WORD currentNumber = GetStringResourceIDNumber(fileName);
      if (currentNumber == idNumber)
      {
        CSubResourceProperties subResourceProps;
        NFileOperationReturnCode::EEnum result = PutStringItem(fileName,
            item.Size, subResourceProps);
        if (result != NFileOperationReturnCode::kSuccess)
          return result;
        structure.Add(subResourceProps);
      }
    }
    for (j = 0; j < m_SubResources.Size(); j++)
    {
      const CSubResource &subResource = m_SubResources[j];
      CItemInfo &itemInfo = m_ItemInfoDatabase.Items[subResource.RecordIndex];
      if (m_ItemInfoDatabase.TypeIDs[itemInfo.TypeIndex] != resourceType)
        continue;
      if (!itemInfo.Name.IsNumber())
        continue;
      if (itemInfo.Name.GetAsNumber() != idNumber)
        continue;
      int k;
      for (k = 0; k < structure.Size(); k++)
        if (structure[k].Name == subResource.SubResourceProperties.Name)
          break;
      if (k != structure.Size())
        continue;
      structure.Add(subResource.SubResourceProperties);
    }

    CDataBlock dataBlock;
    m_EncoderManager.WriteItems(resourceType, structure, dataBlock);
    const BYTE *dataPointer = dataBlock.Data;
    DWORD resourceSize = dataBlock.Size;
    CItemNameID resourceName;
    resourceName.SetAsNumber(idNumber);
    if (!manager.Update(resourceType.GetAsLPCTSTR(),
      resourceName.GetAsLPCTSTR(), langID, (void *)dataPointer, resourceSize))
    {
      g_StartupInfo.ShowMessage(kFarResCannotUpdateResource);
      return NFileOperationReturnCode::kError;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}


NFileOperationReturnCode::EEnum CPlugin::PutItems(CUpdateManager &manager,
    LANGID langID,
    const CItemTypeID &resourceType,
    const CResourceInFileInfoVector &items)
{
  // ShowErrorMessage("PutItems begin");

  if (resourceType.GetAsLPCTSTR() == RT_STRING)
    return PutStringItems(manager, langID, resourceType, items);

  for (int i = 0; i < items.Size(); i++)
  {
    const CResourceInFileInfo &item = items[i];
    DWORD fileSize = item.Size;
    const CSysString &fileName = item.FileName;
    CByteDynamicBuffer buffer;
    buffer.EnsureCapacity(fileSize);

    NFile::NIO::CInFile inFile;
    inFile.Open(fileName);
    UINT32 processedSize;
    if (!inFile.Read(buffer, fileSize, processedSize) || processedSize != fileSize)
    {
      g_StartupInfo.ShowMessage(kFarResFileReadingError);
      return NFileOperationReturnCode::kError;
    }
    const BYTE *dataPointer;
    DWORD resourceSize;
  
    CSysString resourceNameString;
    int dotIndex = fileName.ReverseFind('.');
    CDataBlock dataBlock;
    CItemNameID resourceName;
    CByteDynamicBuffer unicodeBuffer;
    if (dotIndex < 0)
    {
      dataPointer = buffer;
      resourceSize = fileSize;
      resourceName.ParseAndSet(fileName);
    }
    else
    {
      resourceNameString = fileName.Left(dotIndex);
      CSysString extension = fileName.Mid(dotIndex + 1);
      CSubResourcePropertiesVector structure;
      CSubResourceProperties subResourceProps;

      if (InUnicode(resourceType) && fileSize >= 2 && *((wchar_t *)((BYTE *)buffer)) == 0xFEFF)
      {
        subResourceProps.Size = fileSize - 2;
        unicodeBuffer.EnsureCapacity(subResourceProps.Size);
        memcpy(unicodeBuffer, buffer + 2, subResourceProps.Size);
        subResourceProps.Data = unicodeBuffer;
      }
      else if (InNonUnicode(resourceType))
      {
        subResourceProps.Size = fileSize * sizeof(wchar_t);
        unicodeBuffer.EnsureCapacity(subResourceProps.Size);
        if (MultiByteToWideChar(CP_ACP, 0,
            (const char *)((const BYTE *)buffer), fileSize,
            (wchar_t *)((BYTE *)unicodeBuffer), fileSize) != int(fileSize))
        {
          ShowErrorMessage("Internal error #2337");
          return NFileOperationReturnCode::kError;
        }
        subResourceProps.Data = unicodeBuffer;
      }
      else
      {
        subResourceProps.Data = buffer;
        subResourceProps.Size = fileSize;
      }
      subResourceProps.DefaultExtension = extension;
      resourceName.ParseAndSet(resourceNameString);

      subResourceProps.Name = resourceName;
      structure.Add(subResourceProps);

      m_EncoderManager.WriteItems(resourceType, structure, dataBlock);
      dataPointer = dataBlock.Data;
      resourceSize = dataBlock.Size;
    }
    /*
    ShowErrorMessage(CSysString("Lang: ") + DWORDToString(langID));
    ShowErrorMessage(CSysString("PutItem type is number: ") + (resourceType.IsNumber() ? "Y": "N"));
    ShowErrorMessage(CSysString("PutItem type: ") + resourceType.GetAsString());
    ShowErrorMessage(CSysString("PutItem name is number: ") + (resourceName.IsNumber() ? "Y": "N"));
    ShowErrorMessage(CSysString("PutItem name: ") + resourceName.GetAsString());

    ShowErrorMessage(CSysString("Data Size: ") + DWORDToString(resourceSize));
    */


    if (!manager.Update(resourceType.GetAsLPCTSTR(),
        resourceName.GetAsLPCTSTR(), langID, (void *)dataPointer, resourceSize))
    {
      g_StartupInfo.ShowMessage(kFarResCannotUpdateResource);
      return NFileOperationReturnCode::kError;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}

static const char *kFileMustBeLess4GBMessage = "File size must be less than 4GB";

// check typeDirectoryName: maybe it must be converted from OEM to ANSI
NFileOperationReturnCode::EEnum CPlugin::PutTypeItem(CUpdateManager &manager,
    LANGID langID, const CSysString &typeDirectoryName)
{
  CItemTypeID resourceType;
  WORD resourceTypeNumber;
  if (m_StandardTypesNames.GetTypeNumber(typeDirectoryName,
      resourceTypeNumber))
    resourceType.SetAsNumber(resourceTypeNumber);
  else
    resourceType.ParseAndSet(typeDirectoryName);

  // CCurrentDirRestorer currentDirRestorer;
  // ::SetCurrentDirectory(typeDirectoryName);
  NFile::NFind::CEnumerator enumerator(typeDirectoryName + "\\*");
  NFile::NFind::CFileInfo fi;
  CResourceInFileInfoVector items;
  while (enumerator.Next(fi))
  {
    if (fi.Size > ((UInt32)1 << 31))
    {
      g_StartupInfo.ShowMessage(kFileMustBeLess4GBMessage);
      return NFileOperationReturnCode::kError;
    }
    CResourceInFileInfo CResourceInFileInfo;
    CResourceInFileInfo.FileName = fi.Name;
    CResourceInFileInfo.Size = (DWORD)fi.Size;
    CResourceInFileInfo.IsDir = fi.IsDir();
    items.Add(CResourceInFileInfo);
  }
  return PutItems(manager, langID, resourceType, items);
}

static const char *kLangIDMustBeDirectoryMessage = "Lang ID must be presented as directory";
static const char *kLangIDMustBeNumberMessage = "Lang ID must be presented as number from 0 to 65535";
static const char *kResourceGroupMustBeDirectoryMessage = "Resource Group be presented as directory";


NFileOperationReturnCode::EEnum CPlugin::PutLangItems(CUpdateManager &manager,
    struct PluginPanelItem *panelItems, int numItems)
{
  for (int i = 0; i < numItems; i++)
  {
    const PluginPanelItem &pi = panelItems[i];
    if (!NFile::NFind::NAttributes::IsDir(pi.FindData.dwFileAttributes))
    {
      g_StartupInfo.ShowMessage(kLangIDMustBeDirectoryMessage);
      return NFileOperationReturnCode::kError;
    }
    WORD langWORD;
    CSysString itemName = pi.FindData.cFileName;
    if (!StringToUINT16(itemName, langWORD))
    {
      g_StartupInfo.ShowMessage(kLangIDMustBeNumberMessage);
      return NFileOperationReturnCode::kError;
    }
    LANGID langID = langWORD;
    NFile::NFind::CEnumerator enumerator;
    NFile::NFind::CFileInfo fi;
    CResourceInFileInfoVector items;
    while (enumerator.Next(fi))
    {
      if (!fi.IsDir())
      {
        g_StartupInfo.ShowMessage(kResourceGroupMustBeDirectoryMessage);
        return NFileOperationReturnCode::kError;
      }
      NFileOperationReturnCode::EEnum result = PutTypeItem(manager, langID, fi.Name);
      if (result != NFileOperationReturnCode::kSuccess)
        return result;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::PutTypeItems(CUpdateManager &manager,
    struct PluginPanelItem *panelItems, int numItems)
{
  for (int i = 0; i < numItems; i++)
  {
    const PluginPanelItem &pi = panelItems[i];
    if (!NFile::NFind::NAttributes::IsDir(pi.FindData.dwFileAttributes))
    {
      g_StartupInfo.ShowMessage(kResourceGroupMustBeDirectoryMessage);
      return NFileOperationReturnCode::kError;
    }
    NFileOperationReturnCode::EEnum result = PutTypeItem(manager,
        m_ItemInfoDatabase.LangIDs[m_CurrentLangIndex], pi.FindData.cFileName);
    if (result != NFileOperationReturnCode::kSuccess)
      return result;
  }
  return NFileOperationReturnCode::kSuccess;
}

NFileOperationReturnCode::EEnum CPlugin::PutPanelItems(CUpdateManager &manager,
    struct PluginPanelItem *panelItems, int numItems)
{
  // ShowErrorMessage("PutPanelItems begin");
  CResourceInFileInfoVector items;
  for (int i = 0; i < numItems; i++)
  {
    const PluginPanelItem &pi = panelItems[i];
    if (pi.FindData.nFileSizeHigh != 0)
    {
      g_StartupInfo.ShowMessage(kFileMustBeLess4GBMessage);
      return NFileOperationReturnCode::kError;
    }
    CResourceInFileInfo CResourceInFileInfo;
    CResourceInFileInfo.FileName = pi.FindData.cFileName;

    CResourceInFileInfo.Size = pi.FindData.nFileSizeLow;
    CResourceInFileInfo.IsDir = NFile::NFind::NAttributes::IsDir(
        pi.FindData.dwFileAttributes);
    items.Add(CResourceInFileInfo);
  }
  return PutItems(manager, m_ItemInfoDatabase.LangIDs[m_CurrentLangIndex],
      m_ItemInfoDatabase.TypeIDs[m_CurrentTypeIndex], items);
}

NFileOperationReturnCode::EEnum CPlugin::PutFiles(struct PluginPanelItem *panelItems, int numItems,
    int moveMode, int /* opMode */)
{
  if (!IsOSHigherThanNT4())
  {
    g_StartupInfo.ShowMessage(kFarResUpdatingOnlyUnderNT);
    return NFileOperationReturnCode::kError;
  }

  CUpdateManager manager;
  if (!manager.BeginUpdate(m_FileName, false))
  {
    g_StartupInfo.ShowMessage(kFarResCannotBeginResourceUpdate);
    return NFileOperationReturnCode::kError;
  }
  
  NFileOperationReturnCode::EEnum result = NFileOperationReturnCode::kError;
  switch(m_Level)
  {
    case NLevel::kLang: result = PutLangItems(manager, panelItems, numItems); break;
    case NLevel::kType: result = PutTypeItems(manager, panelItems, numItems); break;
    case NLevel::kItems: result = PutPanelItems(manager, panelItems, numItems); break;
  }
  if (result != NFileOperationReturnCode::kSuccess)
    return result;

  if (!manager.EndUpdate(false))
  {
    g_StartupInfo.ShowMessage(kFarResCannotEndResourceUpdate);
    ShowErrorMessage("Could not write changes to file.");
    return NFileOperationReturnCode::kError;
  }
  // ShowErrorMessage("after EndUpdate");
  NormalizeState();
  // m_ResourcesAreLoaded = false;
  if (moveMode != 0)
  {
    for (int i = 0; i < numItems; i++)
    {
      const PluginPanelItem &pluginPanelItem = panelItems[i];
      bool result;
      if (NFile::NFind::NAttributes::IsDir(pluginPanelItem.FindData.dwFileAttributes))
        result = NFile::NDirectory::RemoveDirectoryWithSubItems(pluginPanelItem.FindData.cFileName);
      else
        result = NFile::NDirectory::DeleteFileAlways(pluginPanelItem.FindData.cFileName);
      if (!result)
        return NFileOperationReturnCode::kError;
    }
  }
  return NFileOperationReturnCode::kSuccess;
}

EXTERN_C int WINAPI DeleteFiles(HANDLE hPlugin, PluginPanelItem *panelItems, int numItems, int opMode)
{
  MY_TRY_BEGIN;
  CPlugin *plugin = (CPlugin *)hPlugin;
  return(plugin->DeleteFiles(panelItems , numItems, opMode));
  MY_TRY_END2("DeleteFiles", FALSE);
}

bool CPlugin::DeleteOneRecord(CUpdateManager &manager, const CItemInfo &itemInfo)
{
  if (!manager.Delete(
      m_ItemInfoDatabase.TypeIDs[itemInfo.TypeIndex].GetAsLPCTSTR(),
      itemInfo.Name.GetAsLPCTSTR(),
      MAKELANGID(m_ItemInfoDatabase.LangIDs[itemInfo.LangIndex], SUBLANG_NEUTRAL)))
  {
    g_StartupInfo.ShowMessage(kFarResCannotUpdateResource);
    return false;
  }
  return true;
}

bool CPlugin::DeleteTypeItems(CUpdateManager &manager, int langIndex, int typeIndex)
{
  for (int i = 0; i < m_ItemInfoDatabase.Items.Size(); i++)
  {
    const CItemInfo &itemInfo = m_ItemInfoDatabase.Items[i];
    if (itemInfo.LangIndex != langIndex || itemInfo.TypeIndex != typeIndex)
      continue;
    if (!DeleteOneRecord(manager, itemInfo))
      return false;
  }
  return true;
}

bool CPlugin::DeleteLangItems(CUpdateManager &manager, int langIndex)
{
  CUniqueIntVector typeIndices;
  GetTypeIndexesForOneLang(langIndex, typeIndices);
  for (int i = 0; i < typeIndices.Size(); i++)
    if (!DeleteTypeItems(manager, langIndex, typeIndices[i]))
      return false;
  return true;
}

bool CPlugin::DeleteSubResurceItems(CUpdateManager &manager, CRecordVector<int> &subResourceIndexes)
{
  CUniqueIntVector uniqueRecordIndexes;
  int i;
  for (i = 0; i < subResourceIndexes.Size(); i++)
  {
    const CSubResource &subResource = m_SubResources[subResourceIndexes[i]];
    uniqueRecordIndexes.AddUnique(subResource.RecordIndex);
  }
  for (i = 0; (int)i < uniqueRecordIndexes.Size(); i++)
  {
    int recordIndex = uniqueRecordIndexes[i];
    const CItemInfo &resourceRecord = m_ItemInfoDatabase.Items[recordIndex];
    const CItemTypeID &type = m_ItemInfoDatabase.TypeIDs[resourceRecord.TypeIndex];
    if (type.GetAsLPCTSTR() == RT_STRING)
    {
      CSubResourcePropertiesVector structure;
      for (int j = 0; j < m_SubResources.Size(); j++)
      {
        const CSubResource &subResource = m_SubResources[j];
        if (recordIndex != subResource.RecordIndex)
          continue;
        
        int k;

        for (k = 0; k < subResourceIndexes.Size(); k++)
        {
          if (j == subResourceIndexes[k])
            break;
        }
        if (k != subResourceIndexes.Size())
          continue;
        structure.Add(subResource.SubResourceProperties);
      }
      if (structure.Size() == 0)
      {
        if (!DeleteOneRecord(manager, resourceRecord))
          return false;
      }
      else
      {
        CDataBlock dataBlock;
        m_EncoderManager.WriteItems(type, structure, dataBlock);
        const BYTE *dataPointer = dataBlock.Data;
        DWORD resourceSize = dataBlock.Size;
        if (!manager.Update(type.GetAsLPCTSTR(),
            resourceRecord.Name.GetAsLPCTSTR(),
            m_ItemInfoDatabase.LangIDs[resourceRecord.LangIndex],
            (void *)dataPointer, resourceSize))
        {
          g_StartupInfo.ShowMessage(kFarResCannotUpdateResource);
          return false;
        }
      }
    }
    else
    {
      if (!DeleteOneRecord(manager, resourceRecord))
        return false;
    }
  }
  return true;
}

int CPlugin::DeleteFiles(PluginPanelItem *panelItems, int numItems, int opMode)
{
  if (!IsOSHigherThanNT4())
  {
    g_StartupInfo.ShowMessage(kFarResUpdatingOnlyUnderNT);
    return FALSE;
  }
  if (numItems == 0)
    return FALSE;
  if ((opMode & OPM_SILENT) == 0)
  {
    const char *msgItems[]=
    {
      g_StartupInfo.GetMsgString(kFarResMsgDeleteTitle),
      g_StartupInfo.GetMsgString(kFarResMsgDeleteFiles),
      g_StartupInfo.GetMsgString(kFarResMsgDeleteDelete),
      g_StartupInfo.GetMsgString(kFarResMsgDeleteCancel)
    };
    char msg[1024];
    if (numItems == 1)
    {
      sprintf(msg, g_StartupInfo.GetMsgString(kFarResMsgDeleteFile), panelItems[0].FindData.cFileName);
      msgItems[1] = msg;
    }
    else if (numItems > 1)
    {
      sprintf(msg, g_StartupInfo.GetMsgString(kFarResMsgDeleteNumberOfFiles), numItems);
      msgItems[1] = msg;
    }
    if (g_StartupInfo.ShowMessage(FMSG_WARNING, NULL, msgItems, sizeof(msgItems) / sizeof(msgItems[0]), 2) != 0)
      return FALSE;
  }

  CUpdateManager manager;
  if (!manager.BeginUpdate(m_FileName, false))
  {
    g_StartupInfo.ShowMessage(kFarResCannotBeginResourceUpdate);
    return FALSE;
  }
  
  if (m_Level == NLevel::kItems)
  {
    CRecordVector<int> subResourceIndexes;
    for (int i = 0; i < numItems; i++)
      subResourceIndexes.Add(panelItems[i].UserData);
    if (!DeleteSubResurceItems(manager, subResourceIndexes))
      return FALSE;
  }
  else
    for (int i = 0; i < numItems; i++)
    {
      int index = panelItems[i].UserData;
      // g_StartupInfo.ShowMessage(DWORDToString(index));
      bool result = false;
      switch(m_Level)
      {
        case NLevel::kLang: result = DeleteLangItems(manager, index); break;
        case NLevel::kType: result = DeleteTypeItems(manager, m_CurrentLangIndex, index); break;
      }
      if (!result)
        return FALSE;
    }
    /*
    CSysString itemName = pi.FindData.cFileName;
    if (itemName.Right(4).CompareNoCase(".bmp") != 0)
      break;
    CSysString name = itemName.Left(itemName.Length() - 4);
    if (name.IsEmpty())
    {
      ShowErrorMessage("There is no file name");
      return NFileOperationReturnCode::kError;
    }

    DWORD number;
    bool isNumber = StringToDWORD(name, number);
    LPCTSTR namePointer = isNumber? MAKEINTRESOURCE(number): name;
    ShowErrorMessage("Delete " + name);

    if (!manager.Delete(RT_BITMAP, namePointer,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL)))
    {
      g_StartupInfo.ShowMessage(kFarResCannotUpdateResource);
      return NFileOperationReturnCode::kError;
    }
    */
  if (!manager.EndUpdate(false))
  {
    g_StartupInfo.ShowMessage(kFarResCannotEndResourceUpdate);
    return FALSE;
  }
  NormalizeState();
  // m_ResourcesAreLoaded = false;
  return TRUE;
}


void CPlugin::NormalizeState()
{
  LANGID langId = 0;
  CItemTypeID type;
  if (m_Level != NLevel::kLang)
  {
    langId = m_ItemInfoDatabase.LangIDs[m_CurrentLangIndex];
    if (m_Level == NLevel::kItems)
      type = m_ItemInfoDatabase.TypeIDs[m_CurrentTypeIndex];
  }
  LoadAllResources();
  if (m_Level == NLevel::kLang)
    return;

  int i;
  for (i = 0; i < m_ItemInfoDatabase.LangIDs.Size(); i++)
    if (m_ItemInfoDatabase.LangIDs[i] == langId)
    {
      m_CurrentLangIndex = i;
      break;
    }

  if (i == m_ItemInfoDatabase.LangIDs.Size())
  {
    m_Level = NLevel::kLang;
    m_CurrentDir.Empty();
    return;
  }
  
  if (m_Level != NLevel::kItems)
    return;
  
  for (i = 0; i < m_ItemInfoDatabase.TypeIDs.Size(); i++)
    if (m_ItemInfoDatabase.TypeIDs[i] == type)
    {
      m_CurrentTypeIndex = i;
      return;
    }
  
  SetDirectory("..", 0);
}
