// Windows/Resource/AccleratorCoder.cpp

#include "StdAfx.h"

#include "AcceleratorCoder.h"
#include "CoderUtils.h"

#include "StringSubstitute.h"

using namespace std;

namespace NWindows{
namespace NResource{
namespace NAcclerator{

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharTAB = 9;
static const wchar_t kCharQuote = L'\"';

static const wchar_t kCharSpace = L' ';
static const wchar_t kCharComma = L',';

static const wchar_t *kPropertyDelimiter = L", ";

// static const wchar_t kCharORDelimeter = L'|';
static const wchar_t *kCharORWithSpaceDelimeter = L" | ";

static const wchar_t *kAccleratorInfoKeyString = L"VS_Acclerator_INFO";

static const wchar_t *kValueString = L"VALUE";

static const wchar_t *kStringFileInfoString = L"StringFileInfo";
static const wchar_t *kVarFileInfoString = L"VarFileInfo";

static const wchar_t *kTranslationString = L"Translation";
 
namespace NFixedFileInfo
{
  static const wchar_t *kFileAccleratorString = L"FILEAcclerator";
  static const wchar_t *kProductAccleratorString = L"PRODUCTAcclerator";
  static const wchar_t *kFileFlagsMaskString = L"FILEFLAGSMASK";
  static const wchar_t *kFileFlagsString = L"FILEFLAGS";
  static const wchar_t *kFileOSString = L"FILEOS";
  static const wchar_t *kFileTypeString = L"FILETYPE";
  static const wchar_t *kFileSubTypeString = L"FILESUBTYPE";

  static const UINT32 kSignature = 0xFEEF04BD;
}

static const wchar_t *kBlockString = L"BLOCK";

static const wchar_t *kBeginString = L"{";
static const wchar_t *kEndString = L"}";


static const UINT32 kMaxNumberValue = 0xFFFF;
static const UINT32 kMaxNumDigits = 5;

static bool UStringToWORD(const UString &value, UINT16 &result)
{
  if (value.Length() > kMaxNumDigits)
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  UINT32 dwordResult = wcstoul(value, &endptr, 10);
  result = UINT16(dwordResult);
  return (dwordResult <= kMaxNumberValue);
}

static bool UStringToDWORD(const UString &value, UINT32 &result)
{
  if (value.Length() > 12) // test it !! It was kMaxNumDigits
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  result = wcstoul(value, &endptr, 10);
  return true;
}

UString GetQuotedString(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += s;
  textString += kCharQuote;
  return textString;
}

/*
struct VS_AccleratorINFO {
    WORD  wLength;
    WORD  wValueLength;
    WORD  wType;
    WCHAR szKey[];
    WORD  Padding1[];
    VS_FIXEDFILEINFO Value;
    WORD  Padding2[];
    WORD  Children[];
};
*/

//------------------------------------
// CDecoder

static const char *kAccleratorResourceDefaultExtension = "txt";

void CDecoder::GetProperties(CCoderProperties &aCoderProperties) const
{
  aCoderProperties.AllTypesAreSuported = false;
  // aCoderProperties.SupportedType.Set(RT_Acclerator);
  aCoderProperties.SupportedType.Set(RT_ACCELERATOR);
}

// ----------------------
// CAccleratorStringSubstituter
/*
static const CSubstituteInitPair kAccleratorSubstituteInitPairs[] =
{
  { L"\xA", L"\\n" },
  { L"\xD", L"\\r" },
  { L"\x8", L"\\a" },
  { L"\x9", L"\\t" },
  { L"\"", L"\"\"" },
  { L"\\", L"\\\\" }
};


class CAccleratorStringSubstituter
{
  CStringSubstituter m_Substituter;
public:
  CAccleratorStringSubstituter();
  UString ConvertBinToText(const UString &aBinString);
  UString ConvertTextToBin(const UString &aBinString);
};

CAccleratorStringSubstituter::CAccleratorStringSubstituter()
{
  m_Substituter.Init(kAccleratorSubstituteInitPairs,
    sizeof(kAccleratorSubstituteInitPairs) / sizeof(kAccleratorSubstituteInitPairs[0]));
}

UString CAccleratorStringSubstituter::ConvertBinToText(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += m_Substituter.Substitute(s, true);
  textString += kCharQuote;
  return textString;
}

UString CAccleratorStringSubstituter::ConvertTextToBin(const UString &s)
{
  return m_Substituter.Substitute(s, false);
}

static CAccleratorStringSubstituter g_AccleratorStringSubstituter;
*/
// ----------------------------------------------------------

  
static void AddNewLine(UString &resultString)
{
  resultString += kCharCR;
  resultString += kCharLF;
}

// ----------------------------------------
// TTStandardAcclerator

struct CLangCodePagePair
{
  UINT16 Lang;
  UINT16 CodePage;
};

struct CVar
{
  bool m_IsTextData;
  vector<CLangCodePagePair> m_LangCodePagePairs;
};

struct CVarFileInfo
{
  bool m_IsTextData;
  CObjectVector<CVar> m_Vars;
};

struct CStringPair
{
  bool m_IsTextData;
  UString m_Key;
  UString m_Value;
};

struct CStringTable
{
  bool m_IsTextData;
  UString m_Key;
  CObjectVector<CStringPair> m_StringPairs;
};

struct CStringFileInfo
{
  bool m_IsTextData;
  CObjectVector<CStringTable> m_StringTables;
};

class CInfo
{
  bool m_FixedFileInfoIsDefined;
  bool m_IsTextData;
  VS_FIXEDFILEINFO m_FixedFileInfo;
  vector<bool> m_StringFileInfoIsUsedVector;
  CObjectVector<CStringFileInfo> m_StringFileInfoItems;
  CObjectVector<CVarFileInfo> m_VarFileInfoItems;
public:
  

  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
};

// ----------------------------------------
// CFileOSOptions

class CFileOSOptions: public COptionsCommon
{
public:
  CFileOSOptions()
  {
    AddType(0x00000000, L"VOS_UNKNOWN");
    AddType(0x00010000, L"VOS_DOS");
    AddType(0x00020000, L"VOS_OS216");
    AddType(0x00030000, L"VOS_OS232");
    AddType(0x00040000, L"VOS_NT");
    AddType(0x00000001, L"VOS__WINDOWS16");
    AddType(0x00000002, L"VOS__PM16");
    AddType(0x00000003, L"VOS__PM32");
    AddType(0x00000004, L"VOS__WINDOWS32");
  }
};
CFileOSOptions g_FileOSOptions;

class CFileFlagsOptions: public COptionsCommon
{
public:
  CFileFlagsOptions()
  {
    AddType(0x00000001, L"VS_FF_DEBUG");
    AddType(0x00000002, L"VS_FF_PRERELEASE");
    AddType(0x00000004, L"VS_FF_PATCHED");
    AddType(0x00000008, L"VS_FF_PRIVATEBUILD");
    AddType(0x00000010, L"VS_FF_INFOINFERRED");
    AddType(0x00000020, L"VS_FF_SPECIALBUILD");
  }
};
CFileFlagsOptions g_FileFlagsOptions;

class CFileTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileTypeValue()
  {
    AddType(0x00000000, L"VFT_UNKNOWN");
    AddType(0x00000001, L"VFT_APP");
    AddType(0x00000002, L"VFT_DLL");
    AddType(0x00000003, L"VFT_DRV");
    AddType(0x00000004, L"VFT_FONT");
    AddType(0x00000005, L"VFT_VXD");
    AddType(0x00000007, L"VFT_STATIC_LIB");
  }
};
CFileTypeValue g_FileTypeValue;

class CFileDriverSubTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileDriverSubTypeValue()
  {
    AddType(0x00000000, L"VFT2_UNKNOWN");
    AddType(0x00000001, L"VFT2_DRV_PRINTER");
    AddType(0x00000002, L"VFT2_DRV_KEYBOARD");
    AddType(0x00000003, L"VFT2_DRV_LANGUAGE");
    AddType(0x00000004, L"VFT2_DRV_DISPLAY");
    AddType(0x00000005, L"VFT2_DRV_MOUSE");
    AddType(0x00000006, L"VFT2_DRV_NETWORK");
    AddType(0x00000007, L"VFT2_DRV_SYSTEM");
    AddType(0x00000008, L"VFT2_DRV_INSTALLABLE");
    AddType(0x00000009, L"VFT2_DRV_SOUND");
    AddType(0x0000000A, L"VFT2_DRV_COMM");
    AddType(0x0000000B, L"VFT2_DRV_INPUTMETHOD");
  }
};
CFileDriverSubTypeValue g_FileDriverSubTypeValue;

class CFileFontSubTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileFontSubTypeValue()
  {
    AddType(0x00000000, L"VFT2_UNKNOWN");
    AddType(0x00000001, L"VFT2_FONT_RASTER");
    AddType(0x00000002, L"VFT2_FONT_VECTOR");
    AddType(0x00000003, L"VFT2_FONT_TRUETYPE");
  }
};
CFileFontSubTypeValue g_FileFontSubTypeValue;
// -----------------------------
// Load from Binary

static bool GetBoolFromWord(UINT16 word)
{
  switch(word)
  {
    case 0:
      return false;
    case 1:
      return true;
    default:
      throw CDecoderException();
  }
}

void CInfo::LoadFromBinary(CDataInStream &inStream)
{
  UINT16 aLength = inStream.GetWORD();
  UINT16 valueLength = inStream.GetWORD();
  int y = sizeof(VS_FIXEDFILEINFO);
  if (valueLength == 0)
    m_FixedFileInfoIsDefined = false;
  else if (valueLength == sizeof(VS_FIXEDFILEINFO))
    m_FixedFileInfoIsDefined = true;
  else
    throw CDecoderException();
  m_IsTextData = GetBoolFromWord(inStream.GetWORD());
  if (m_FixedFileInfoIsDefined)
  {
    int aKeyLength = (wcslen(kAccleratorInfoKeyString) + 1) * sizeof(wchar_t);
    if (wcscmp(inStream.GetData(), kAccleratorInfoKeyString) != 0)
      throw CDecoderException();
    inStream.SkipBytes(aKeyLength);
  }
  inStream.SkipToAlignDWORD();
  memmove(&m_FixedFileInfo, inStream.GetData(), sizeof(m_FixedFileInfo));
  inStream.SkipBytes(sizeof(m_FixedFileInfo));
  while (inStream.GetBytePosition() < aLength)
  {
    inStream.SkipToAlignDWORD();
    UINT16 anItemEndPosition = inStream.GetBytePosition();
    UINT16 anItemLength = inStream.GetWORD();
    anItemEndPosition += anItemLength;

    UINT16 valueLength = inStream.GetWORD();
    if (valueLength != 0)
      throw CDecoderException();
    bool anIsTextData = GetBoolFromWord(inStream.GetWORD());;
    UString aKey = inStream.GetString();
    inStream.SkipToAlignDWORD();
    if (aKey == kVarFileInfoString)
    {
      CVarFileInfo aVarFileInfo;
      aVarFileInfo.m_IsTextData = anIsTextData;
      inStream.SkipToAlignDWORD();
      while (inStream.GetBytePosition() < anItemEndPosition)
      {
        inStream.SkipToAlignDWORD();
        CVar aVar;
        inStream.GetWORD(); // wLength
        UINT16 valueLength = inStream.GetWORD(); // wLength
        int aNumItems = valueLength / sizeof(UINT32);
        aVar.m_IsTextData = GetBoolFromWord(inStream.GetWORD());;
        UString aKey = inStream.GetString();
        if (aKey != kTranslationString)
          throw CDecoderException();
        inStream.SkipToAlignDWORD();
        for (int i = 0; i < aNumItems; i++)
        {
          CLangCodePagePair aLangCodePagePair;
          UINT32 aDWORD = inStream.GetDWORD();
          aLangCodePagePair.Lang = LOWORD(aDWORD);
          aLangCodePagePair.CodePage = HIWORD(aDWORD);
          aVar.m_LangCodePagePairs.push_back(aLangCodePagePair);
        }
        aVarFileInfo.m_Vars.Add(aVar);
      }
      m_VarFileInfoItems.Add(aVarFileInfo);
      m_StringFileInfoIsUsedVector.push_back(false);
    }
    else if (aKey == kStringFileInfoString)
    {
      CStringFileInfo sFileInfo;
      sFileInfo.m_IsTextData = anIsTextData;
      inStream.SkipToAlignDWORD();
      while (inStream.GetBytePosition() < anItemEndPosition)
      {
        inStream.SkipToAlignDWORD();
        CStringTable sTable;
        UINT16 anItemEndPositionofStringTable = inStream.GetBytePosition();
        UINT16 aLength = inStream.GetWORD();
        anItemEndPositionofStringTable += aLength;
        inStream.GetWORD(); // valueLength
        sTable.m_IsTextData = GetBoolFromWord(inStream.GetWORD());
        sTable.m_Key = inStream.GetString();
        inStream.SkipToAlignDWORD();
        while (inStream.GetBytePosition() < anItemEndPositionofStringTable)
        {
          inStream.SkipToAlignDWORD();
          CStringPair sPair;
          inStream.GetWORD(); // Length
          UINT16 valueLength = inStream.GetWORD();
          sPair.m_IsTextData = GetBoolFromWord(inStream.GetWORD());
          sPair.m_Key = inStream.GetString();
          inStream.SkipToAlignDWORD();
          if (valueLength > 0)
            sPair.m_Value = inStream.GetString();
          sTable.m_StringPairs.Add(sPair);
        }
        sFileInfo.m_StringTables.Add(sTable);
      }
      m_StringFileInfoItems.Add(sFileInfo);
      m_StringFileInfoIsUsedVector.push_back(true);
    }
    else
      throw CDecoderException();
  }
}


// ----------------------------
// Save to Text

UString GetAccleratorString(DWORD aMS, DWORD aLS)
{
  UString s;
  s += UnsignedToUString(HIWORD(aMS));
  s += kCharComma;
  s += UnsignedToUString(LOWORD(aMS));
  s += kCharComma;
  s += UnsignedToUString(HIWORD(aLS));
  s += kCharComma;
  s += UnsignedToUString(LOWORD(aLS));
  return s;
}

UString GetOptions(const COptionsCommon &anOptionsCommon, UINT32 aFlags)
{
  UString result;
  if (aFlags == 0)
  {
    if (anOptionsCommon.GetFirstZeroFlagName(result))
      return result;
    return UnsignedToHexUString(aFlags);
  }
  UINT32 notUsedFlags;
  UStringVector names;
  anOptionsCommon.GetOptionNames(aFlags, names, notUsedFlags);
  for (int i = 0; i < names.Size(); i++)
  {
    if (i != 0)
      result += kCharORWithSpaceDelimeter;
    result += names[i];
  }
  if (notUsedFlags != 0)
  {
    if (names.Size() > 0)
      result += kCharORWithSpaceDelimeter;
    result += UnsignedToHexUString(aFlags);
  }
  return result;
}

UString GetValue(const CUINT32FromToUStringMap &aMap, UINT32 number)
{
  UString s;
  if (aMap.GetString(number, s))
    return s;
  else
    return UnsignedToUString(number);
}

static UString GetStringWithSpaces(const UString &s, int size)
{
  UString result = s;
  while (result.Length() < size)
    result += kCharSpace;
  result += kCharSpace;
  return result;
}

static UString GetFixedInfoPropertyCaption(const UString &s)
{
  return GetStringWithSpaces(s, 14);
}

void CInfo::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  if (m_FixedFileInfoIsDefined)
  {
    if (m_FixedFileInfo.dwSignature != NFixedFileInfo::kSignature)
      throw CDecoderException();
    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileAccleratorString);
    outStream += GetAccleratorString(m_FixedFileInfo.dwFileAccleratorMS,
        m_FixedFileInfo.dwFileAccleratorLS);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kProductAccleratorString);
    outStream += GetAccleratorString(m_FixedFileInfo.dwProductAccleratorMS,
        m_FixedFileInfo.dwProductAccleratorLS);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileFlagsMaskString);
    outStream += UnsignedToHexUString(m_FixedFileInfo.dwFileFlagsMask);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileFlagsString);
    outStream += GetOptions(g_FileFlagsOptions, m_FixedFileInfo.dwFileFlags);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileOSString);
    outStream += GetOptions(g_FileOSOptions, m_FixedFileInfo.dwFileOS);
    AddNewLine(outStream);
  
    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileTypeString);
    outStream += GetValue(g_FileTypeValue, m_FixedFileInfo.dwFileType);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileSubTypeString);
    switch(m_FixedFileInfo.dwFileType)
    {
      case VFT_DRV:
        outStream += GetValue(g_FileDriverSubTypeValue, m_FixedFileInfo.dwFileSubtype);
        break;
      case VFT_FONT:
        outStream += GetValue(g_FileFontSubTypeValue, m_FixedFileInfo.dwFileSubtype);
        break;
      default:
        outStream += UnsignedToHexUString(m_FixedFileInfo.dwFileSubtype);
    }
    AddNewLine(outStream);
  }
  int anStringIndex = 0, aVarIndex = 0;
  outStream += kBeginString;
  indentMaker.AddIndent();
  for (int i = 0; i < m_StringFileInfoIsUsedVector.size(); i++)
  {
    AddNewLine(outStream);
    indentMaker.WriteIndent(outStream);
    if (m_StringFileInfoIsUsedVector[i]) // StringTable
    {
      const CStringFileInfo sFileInfo = m_StringFileInfoItems[anStringIndex++];
      outStream += kBlockString;
      outStream += kCharSpace;
      outStream += GetQuotedString(kStringFileInfoString);
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kBeginString;
      indentMaker.AddIndent();
      for (int j = 0; j < sFileInfo.m_StringTables.Size(); j++)
      {
        const CStringTable &sTable = sFileInfo.m_StringTables[j];
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kBlockString;
        outStream += kCharSpace;
        outStream += GetQuotedString(sTable.m_Key);
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kBeginString;
        indentMaker.AddIndent();
        for (int k = 0; k < sTable.m_StringPairs.Size(); k++)
        {
          AddNewLine(outStream);
          indentMaker.WriteIndent(outStream);
          outStream += kValueString;
          outStream += kCharSpace;
          const CStringPair &sPair = sTable.m_StringPairs[k];
          outStream += GetStringWithSpaces(GetQuotedString(sPair.m_Key) +
              kCharComma, 20);
          outStream += GetQuotedString(sPair.m_Value);
        }
        indentMaker.DeleteIndent();
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kEndString;
      }
      indentMaker.DeleteIndent();
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kEndString;
    }
    else
    {
      const CVarFileInfo aVarFileInfo = m_VarFileInfoItems[aVarIndex++];
      outStream += kBlockString;
      outStream += kCharSpace;
      outStream += GetQuotedString(kVarFileInfoString);
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kBeginString;
      indentMaker.AddIndent();
      for (int j = 0; j < aVarFileInfo.m_Vars.Size(); j++)
      {
        const CVar &aVar = aVarFileInfo.m_Vars[j];
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kValueString;
        outStream += kCharSpace;
        outStream += GetQuotedString(kTranslationString);
        outStream += kPropertyDelimiter;
        for (int k = 0; k < aVar.m_LangCodePagePairs.size(); k++)
        {
          const CLangCodePagePair &aLangCodePagePair = aVar.m_LangCodePagePairs[k];
          if (k > 0)
          {
            outStream += kPropertyDelimiter;
            outStream += kCharSpace;
          }
          outStream += UnsignedToHexUString(aLangCodePagePair.Lang);
          outStream += kPropertyDelimiter;
          outStream += UnsignedToUString(aLangCodePagePair.CodePage);
        }
      }
      indentMaker.DeleteIndent();
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kEndString;
    }
  }
  indentMaker.DeleteIndent();
  AddNewLine(outStream);
  indentMaker.WriteIndent(outStream);
  outStream += kEndString;


  // indentMaker.WriteIndent(outStream);
}


//------------------------------------
// CDecoder

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = kAccleratorResourceDefaultExtension;

  UINT32 resourceSize = resourceData.Size / 2;
  UINT16 *pData = (UINT16 *)((BYTE *)resourceData.Data);
  CDataInStream aStream(pData, resourceSize);
  UINT16 aLength = aStream.PeekWORD(0);
  if (aLength > resourceData.Size)
    throw CDecoderException();

  CIndentMaker indentMaker;
  UString outStream;
  CInfo anInfo;
  anInfo.LoadFromBinary(aStream);
  anInfo.SaveToText(outStream, indentMaker);
  resourceProps.Size = outStream.Length() * sizeof(wchar_t);
  resourceProps.Data.EnsureCapacity(resourceProps.Size);
  memmove(resourceProps.Data, (LPCWSTR)outStream, resourceProps.Size);
}


}}}
