// ResBitmap.cpp

#include "StdAfx.h"

#include "Common/IntToString.h"

#include "CoderUtils.h"
#include "ResBitmap.h"

namespace NWindows {
namespace NResource {
namespace NBitmap {

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_BITMAP);
}

static const TCHAR kDelimiterChar = TEXT('*');

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("bmp");
  
  BITMAPFILEHEADER header;
  BITMAPINFOHEADER &info = *((BITMAPINFOHEADER *)resourceData.Data);
  if (info.biSize != sizeof(info))
    throw CDecoderException();
  
  UInt32 resourceSize = resourceData.Size;
  UInt32 imageSize = info.biSizeImage;
  if (imageSize == 0)
  {
    if (info.biCompression != BI_RGB)
      throw CDecoderException();
    DWORD lineSize = ((info.biWidth * info.biBitCount + 7) / 8 + 3) / 4 * 4;
    LONG ySize = info.biHeight;
    if (ySize < 0)
      ySize = -ySize;
    imageSize = lineSize * ySize;
  }
  header.bfOffBits = sizeof(header) + resourceSize - imageSize;
  
  resourceProps.Size = sizeof(header) + resourceSize;
  resourceProps.Description = DWORDToString(info.biWidth) +
      kDelimiterChar + DWORDToString(info.biHeight) +
      kDelimiterChar + DWORDToString(info.biBitCount);

  header.bfType = 0x4D42;
  header.bfSize = resourceProps.Size;
  header.bfReserved1 = 0;
  header.bfReserved2 = 0;

  CByteDynamicBuffer &buf = resourceProps.Data;
  buf.SetCapacity(resourceProps.Size);
  memcpy(buf, &header, sizeof(header));
  memcpy(buf+ sizeof(header), resourceData.Data, resourceSize);
}


void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_BITMAP);
}

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock &resourceData) const
{
  if (structure.Size() != 1)
    throw 2053;
  const CSubResourceProperties &subResourceProps = structure[0];
  if (subResourceProps.Size < sizeof(BITMAPFILEHEADER))
    throw 2054;
  DWORD size = subResourceProps.Size - sizeof(BITMAPFILEHEADER);
  resourceData.Data.EnsureCapacity(size);
  resourceData.Size = size;
  memcpy(resourceData.Data, subResourceProps.Data + sizeof(BITMAPFILEHEADER), size);
}

}}}
