// ResCoder.cpp

#include "StdAfx.h"

#include "ResCoder.h"

namespace NWindows {
namespace NResource {

//-------------------------------------
// CDecoderManager

CSysString CSubResourceProperties::GetFullName() const
{
  return DefaultExtension.IsEmpty() ?
      Name.GetAsString() : (Name.GetAsString() + '.' + DefaultExtension);
}

void CDecoderManager::AddDecoder(CDecoder *decoder)
{
  m_Decoders.Add(decoder);
}

void CDecoderManager::ReadItems(const CItemInfoDatabase &db,
    int resourceIndex, const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  for (int i = 0; i < m_Decoders.Size(); i++)
  {
    const CDecoder &decoder = *((const CDecoder *)m_Decoders[i]);
    CCoderProperties decoderProps;
    decoder.GetProperties(decoderProps);
    const CItemInfo &itemInfo = db.Items[resourceIndex];
    if (decoderProps.AllTypesAreSuported || decoderProps.SupportedType == db.TypeIDs[itemInfo.TypeIndex])
    {
      try
      {
        decoder.ReadItems(itemInfo.Name, resourceData, structure);
      }
      catch(const CDecoderException &)
      {
        structure.Clear();
        continue;
      }
      catch(...)
      {
        structure.Clear();
        continue;
      }
      return;
    }
  }
  throw "Unsupported resource type";
}


//------------------------------------
// CDefaultDecoder

void CDefaultDecoder::GetProperties(CCoderProperties &decoderProps) const
{
  decoderProps.AllTypesAreSuported = true;
}

void CDefaultDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &props = structure.Back();
  
  props.Name = resourceName;
  props.Size = resourceData.Size; // Test it;

  CByteDynamicBuffer &byteBuffer = props.Data;
  byteBuffer.EnsureCapacity(props.Size);
  memcpy(byteBuffer, resourceData.Data, props.Size);
}

// Write to resource file
// PackResource

// ------------------------------------------------
// CEncoderManager

void CEncoderManager::AddEncoder(CEncoder *encoder)
{
  m_Encoders.Add(encoder);
}

void CEncoderManager::WriteItems(const CItemTypeID &resourceType,
    const CSubResourcePropertiesVector &structure, CDataBlock &resourceData) const
{
  for (int i = 0; i < m_Encoders.Size(); i++)
  {
    const CEncoder &encoder = *((const CEncoder *)m_Encoders[i]);
    CCoderProperties decoderProps;
    encoder.GetProperties(decoderProps);
    if (decoderProps.AllTypesAreSuported || decoderProps.SupportedType == resourceType)
    {
      encoder.WriteItems(structure, resourceData);
      return;
    }
  }
  throw "Unsupported resource type";
}

void CDefaultEncoder::GetProperties(CCoderProperties &decoderProps) const
{
  decoderProps.AllTypesAreSuported = true;
}

void CDefaultEncoder::WriteItems(const CSubResourcePropertiesVector &structure, CDataBlock &resourceData) const
{
  if (structure.Size() != 1)
    throw 2043;
  const CSubResourceProperties &subResourceProps = structure[0];
  DWORD size = subResourceProps.Size;
  resourceData.Data.EnsureCapacity(size);
  memcpy(resourceData.Data, subResourceProps.Data, size);
}

}}
