// ResCoder.h

#ifndef __WINDOWS_RESOURCE_CODER_H
#define __WINDOWS_RESOURCE_CODER_H

#include "Common/DynamicBuffer.h"

#include "ResUtils.h"

namespace NWindows {
namespace NResource {

struct CDecoderException
{
  CDecoderException() {};
};

struct CCoderProperties
{
  bool AllTypesAreSuported;
  CItemTypeID SupportedType;
};

struct CSubResourceProperties
{
  CItemNameID Name;
  CSysString DefaultExtension;
  CSysString Description;
  UInt32 Size;
  CByteDynamicBuffer Data;
  CSysString GetFullName() const;
};

typedef CObjectVector<CSubResourceProperties> CSubResourcePropertiesVector;

struct CDecoder
{
  virtual void GetProperties(CCoderProperties &props) const = 0;
  virtual void ReadItems(const CItemNameID &resourceName,
      const CDataInfo &dataInfo, CSubResourcePropertiesVector &structure) const = 0;
};

class CDecoderManager
{
  CPointerVector m_Decoders;
public:
  void AddDecoder(CDecoder *aDecoder);
  void ReadItems(const CItemInfoDatabase &db,
      int aResourceIndex,
      const CDataInfo &resourceData,
      CSubResourcePropertiesVector &structure) const;
};

struct CDefaultDecoder: public CDecoder
{
  virtual void GetProperties(CCoderProperties &props) const;
  virtual void ReadItems(const CItemNameID &resourceName,
      const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const;
};


struct CDataBlock
{
  CByteDynamicBuffer Data;
  UINT32 Size;
};

struct CEncoder
{
  virtual void GetProperties(CCoderProperties &props) const = 0;
  virtual void WriteItems(const CSubResourcePropertiesVector &structure,
      CDataBlock &resourceData) const = 0;
};

class CEncoderManager
{
  CPointerVector m_Encoders;
public:
  void AddEncoder(CEncoder *anEncoder);
  void WriteItems(const CItemTypeID &aResourceType,
      const CSubResourcePropertiesVector &structure,
      CDataBlock &resourceData) const;
};

struct CDefaultEncoder: public CEncoder
{
  virtual void GetProperties(CCoderProperties &props) const;
  virtual void WriteItems(const CSubResourcePropertiesVector &structure,
      CDataBlock &resourceData) const;
};

}}

#define MY_RES_DECODER_DECL \
struct CDecoder: public NResource::CDecoder \
{ \
  virtual void GetProperties(CCoderProperties &props) const; \
  virtual void ReadItems(const CItemNameID &resourceName, \
      const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const; \
};

#define MY_RES_ENCODER_DECL \
struct CEncoder: public NResource::CEncoder \
{ \
  virtual void GetProperties(CCoderProperties &props) const; \
  virtual void WriteItems(const CSubResourcePropertiesVector &structure, \
      CDataBlock &resourceData) const; \
};

#endif
