// ResGroupIcon.cpp

#include "StdAfx.h"

#include "Common/StringConvert.h"

#include "CoderUtils.h"
#include "IconCursor.h"
#include "ResGroupIcon.h"

namespace NWindows {
namespace NResource {

using namespace NIconCursor;

static const char kDelimiterChar = '*';

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharSpace = L' ';

static void AddNewLine(UString &resultString)
{
  resultString += kCharCR;
  resultString += kCharLF;
}

static UString UnsignedToUString(UINT32 value, int digits)
{
  wchar_t format[32];
  swprintf(format, L"%%%dd", digits);
  wchar_t tmp[32];
  swprintf(tmp, format, value);
  return tmp;
}

static void ReadCommonItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,CSubResourcePropertiesVector &structure, bool icon)
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("txt");

  const BYTE *data = (const BYTE *)resourceData.Data;
  UINT32 leaveSize = resourceData.Size;

  if (leaveSize < sizeof(CHeader))
    throw CDecoderException();
  
  const CHeader &iconHeader = *((const CHeader *)data);
  leaveSize -= sizeof(CHeader);
  data += sizeof(CHeader);

  if (iconHeader.Reserved != 0)
    throw CDecoderException();
  if (icon)
  {
    if (iconHeader.ResType != RES_ICON)
      throw CDecoderException();
  }
  else
  {
    if (iconHeader.ResType != RES_CURSOR)
      throw CDecoderException();
  }

  UString outStream;
  UString description;

  for (int i = 0; i < iconHeader.ResCount; i++)
  {
    if (i != 0)
      AddNewLine(outStream);
    if (leaveSize < sizeof(CResDirInGroup ))
      throw CDecoderException();
    const CResDirInGroup &resDir = *((const CResDirInGroup  *)data);
    leaveSize -= sizeof(CResDirInGroup );
    data += sizeof(CResDirInGroup );
    
    outStream += UnsignedToUString(resDir.ID, 3);
    outStream += kCharSpace;
    outStream += kCharSpace;

    if (i != 0)
      description += kCharSpace;
    description += UnsignedToUString(resDir.ID, 3);

    if (icon)
    {
      outStream += UnsignedToUString(resDir.Icon.Width, 2);
      outStream += kCharSpace;
      outStream += UnsignedToUString(resDir.Icon.Height, 2);
      outStream += kCharSpace;
      outStream += kCharSpace;
      
      if (resDir.Icon.Reserved != 0)
        throw CDecoderException();
      
      outStream += UnsignedToUString(resDir.Icon.ColorCount, 2);
      outStream += kCharSpace;
    }
    else
    {
      outStream += UnsignedToUString(resDir.Cursor.Width, 2);
      outStream += kCharSpace;
      if ((resDir.Cursor.Height & 1) != 0)
        throw CDecoderException();
      outStream += UnsignedToUString(resDir.Cursor.Height / 2, 2);
      outStream += kCharSpace;
      outStream += kCharSpace;
    }

    if (resDir.Planes != 1) // Planes
      throw CDecoderException();

    outStream += UnsignedToUString(resDir.BitCount, 2); // BitCount
    outStream += kCharSpace;

    outStream += UnsignedToUString(resDir.BytesInRes, 5);
  }
  if (leaveSize != 0)
    throw CDecoderException();
  resourceProps.Description = GetSystemString(description);

  CByteDynamicBuffer &buf = resourceProps.Data;

  resourceProps.Size = outStream.Length() * sizeof(wchar_t);
  buf.EnsureCapacity(resourceProps.Size);
  memcpy(buf, (const wchar_t *)outStream, resourceProps.Size);
}


namespace NGroupIcon {

static const char *kIsNotIconMessage = "Bad ICON Group data";

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_GROUP_ICON);
}

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  ReadCommonItems(resourceName, resourceData, structure, true);
}

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_GROUP_ICON);
}


void CEncoder::WriteItems(const CSubResourcePropertiesVector & /* structure */,
    CDataBlock & /* resourceData */) const
{}

}

namespace NGroupCursor {

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_GROUP_CURSOR);
}

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  ReadCommonItems(resourceName, resourceData, structure, false);
}

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_GROUP_CURSOR);
}

void CEncoder::WriteItems(const CSubResourcePropertiesVector & /* structure */,CDataBlock & /* resourceData */) const
{}

}

}}
