// Windows/Resource/Utils.h

#ifndef __WINDOWS_RESOURCE_UTILS_H
#define __WINDOWS_RESOURCE_UTILS_H

#include "Common/MyString.h"
#include "Common/Types.h"

#include "Windows/DLL.h"

namespace NWindows {
namespace NResource {

bool StringToUINT16(const CSysString &value, UINT16 &result);

class CNumberOrString2
{
  bool m_IsNumber;
};

class CNumberOrString
{
  friend bool operator<(const CNumberOrString &a1, const CNumberOrString &a2);
  friend bool operator==(const CNumberOrString &a1, const CNumberOrString &a2);
  bool m_IsNumber;
  UINT16 m_Number;
  CSysString m_String;
public:
  bool IsNumber() const { return m_IsNumber; };
  void Set(LPCTSTR s);
  void SetAsNumber(UINT16 number);
  void SetAsName(LPCTSTR s);
  void ParseAndSet(LPCTSTR s);
  CSysString GetAsString() const;
  UINT16 GetAsNumber() const;
  LPCTSTR GetAsLPCTSTR() const;
};

bool operator<(const CNumberOrString &a1, const CNumberOrString &a2);
bool operator==(const CNumberOrString &a1, const CNumberOrString &a2);
bool operator!=(const CNumberOrString &a1, const CNumberOrString &a2);

typedef CNumberOrString CItemTypeID;
typedef CNumberOrString CItemNameID;

struct CItemInfo
{
  CItemNameID Name;
  int LangIndex;
  int TypeIndex;
};

struct CItemInfoDatabase
{
  CRecordVector<LANGID> LangIDs;
  CObjectVector<CItemTypeID> TypeIDs;
  CObjectVector<CItemInfo> Items;

  void Clear();
};

struct CDataInfo
{
  LPVOID Data;
  DWORD Size;
};

class CModule: public NDLL::CLibrary
{
public:
  bool ReadItemInfoDatabase(CItemInfoDatabase &db) const;
  bool ReadResource(CDataInfo &dataInfo,
      const CItemInfoDatabase &db, int index) const;
};

 
class CStandardTypesNames
{
  // StandardResourceTypePairList m_Pairs;
  std::map<UINT16, CSysString> m_NumberToStringMap;
  std::map<CSysString, UINT16> m_StringToNumberMap;
public:
  void AddEnglishNames();
  void AddType(LPCTSTR number, const CSysString &name);
  bool GetTypeName(UINT16 number, CSysString &name) const;
  bool GetTypeNumber(const CSysString &name, UINT16 &number) const;
};

}}

#endif
