// ResVersion.cpp

#include "StdAfx.h"

#include "CoderUtils.h"
#include "ResVersion.h"
#include "StringSubstitute.h"

using namespace std;

namespace NWindows {
namespace NResource {
namespace NVersion {

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharTAB = 9;
static const wchar_t kCharQuote = L'\"';

static const wchar_t kCharSpace = L' ';
static const wchar_t kCharComma = L',';

static const wchar_t *kPropertyDelimiter = L", ";

static const wchar_t *kCharORWithSpaceDelimeter = L" | ";

static const wchar_t *kVersionInfoKeyString = L"VS_VERSION_INFO";

static const wchar_t *kValueString = L"VALUE";

static const wchar_t *kStringFileInfoString = L"StringFileInfo";
static const wchar_t *kVarFileInfoString = L"VarFileInfo";

static const wchar_t *kTranslationString = L"Translation";
 
namespace NFixedFileInfo
{
  static const wchar_t *kFileVersionString = L"FILEVERSION";
  static const wchar_t *kProductVersionString = L"PRODUCTVERSION";
  static const wchar_t *kFileFlagsMaskString = L"FILEFLAGSMASK";
  static const wchar_t *kFileFlagsString = L"FILEFLAGS";
  static const wchar_t *kFileOSString = L"FILEOS";
  static const wchar_t *kFileTypeString = L"FILETYPE";
  static const wchar_t *kFileSubTypeString = L"FILESUBTYPE";

  static const UINT32 kSignature = 0xFEEF04BD;
}

static const wchar_t *kBlockString = L"BLOCK";

static const wchar_t *kBeginString = L"{";
static const wchar_t *kEndString = L"}";


static const UINT32 kMaxNumberValue = 0xFFFF;
static const UINT32 kMaxNumDigits = 5;

/*
static bool UStringToWORD(const UString &value, UINT16 &result)
{
  if (value.Length() > kMaxNumDigits)
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  UINT32 dwordResult = wcstoul(value, &endptr, 10);
  result = UINT16(dwordResult);
  return (dwordResult <= kMaxNumberValue);
}

static bool UStringToDWORD(const UString &value, UINT32 &result)
{
  if (value.Length() > 12) // test it !! It was kMaxNumDigits
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  result = wcstoul(value, &endptr, 10);
  return true;
}
*/

UString GetQuotedString(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += s;
  textString += kCharQuote;
  return textString;
}

/*
struct VS_VERSIONINFO {
    WORD  wLength;
    WORD  wValueLength;
    WORD  wType;
    WCHAR szKey[];
    WORD  Padding1[];
    VS_FIXEDFILEINFO Value;
    WORD  Padding2[];
    WORD  Children[];
};
*/

//------------------------------------
// CDecoder

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_VERSION);
}

// ----------------------
// CVersionStringSubstituter
/*
static const CSubstituteInitPair kVersionSubstituteInitPairs[] =
{
  { L"\xA", L"\\n" },
  { L"\xD", L"\\r" },
  { L"\x8", L"\\a" },
  { L"\x9", L"\\t" },
  { L"\"", L"\"\"" },
  { L"\\", L"\\\\" }
};


class CVersionStringSubstituter
{
  CStringSubstituter m_Substituter;
public:
  CVersionStringSubstituter();
  UString ConvertBinToText(const UString &aBinString);
  UString ConvertTextToBin(const UString &aBinString);
};

CVersionStringSubstituter::CVersionStringSubstituter()
{
  m_Substituter.Init(kVersionSubstituteInitPairs,
    sizeof(kVersionSubstituteInitPairs) / sizeof(kVersionSubstituteInitPairs[0]));
}

UString CVersionStringSubstituter::ConvertBinToText(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += m_Substituter.Substitute(s, true);
  textString += kCharQuote;
  return textString;
}

UString CVersionStringSubstituter::ConvertTextToBin(const UString &s)
{
  return m_Substituter.Substitute(s, false);
}

static CVersionStringSubstituter g_VersionStringSubstituter;
*/
// ----------------------------------------------------------

  
static void AddNewLine(UString &resultString)
{
  resultString += kCharCR;
  resultString += kCharLF;
}

struct CLangCodePagePair
{
  UINT16 Lang;
  UINT16 CodePage;
};

struct CVar
{
  bool m_IsTextData;
  CRecordVector<CLangCodePagePair> m_LangCodePagePairs;
};

struct CVarFileInfo
{
  bool m_IsTextData;
  CObjectVector<CVar> m_Vars;
};

struct CStringPair
{
  bool m_IsTextData;
  UString m_Key;
  UString m_Value;
};

struct CStringTable
{
  bool m_IsTextData;
  UString m_Key;
  CObjectVector<CStringPair> m_StringPairs;
};

struct CStringFileInfo
{
  bool m_IsTextData;
  CObjectVector<CStringTable> m_StringTables;
};

class CInfo
{
  bool m_FixedFileInfoIsDefined;
  bool m_IsTextData;
  VS_FIXEDFILEINFO m_FixedFileInfo;
  CBoolVector m_StringFileInfoIsUsedVector;
  CObjectVector<CStringFileInfo> m_StringFileInfoItems;
  CObjectVector<CVarFileInfo> m_VarFileInfoItems;
public:
  

  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
};

// ----------------------------------------
// CFileOSOptions

class CFileOSOptions: public COptionsCommon
{
public:
  CFileOSOptions()
  {
    AddType(0x00000000, L"VOS_UNKNOWN");
    AddType(0x00010000, L"VOS_DOS");
    AddType(0x00020000, L"VOS_OS216");
    AddType(0x00030000, L"VOS_OS232");
    AddType(0x00040000, L"VOS_NT");
    AddType(0x00000001, L"VOS__WINDOWS16");
    AddType(0x00000002, L"VOS__PM16");
    AddType(0x00000003, L"VOS__PM32");
    AddType(0x00000004, L"VOS__WINDOWS32");
  }
};
CFileOSOptions g_FileOSOptions;

class CFileFlagsOptions: public COptionsCommon
{
public:
  CFileFlagsOptions()
  {
    AddType(0x00000001, L"VS_FF_DEBUG");
    AddType(0x00000002, L"VS_FF_PRERELEASE");
    AddType(0x00000004, L"VS_FF_PATCHED");
    AddType(0x00000008, L"VS_FF_PRIVATEBUILD");
    AddType(0x00000010, L"VS_FF_INFOINFERRED");
    AddType(0x00000020, L"VS_FF_SPECIALBUILD");
  }
};
CFileFlagsOptions g_FileFlagsOptions;

class CFileTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileTypeValue()
  {
    AddType(0x00000000, L"VFT_UNKNOWN");
    AddType(0x00000001, L"VFT_APP");
    AddType(0x00000002, L"VFT_DLL");
    AddType(0x00000003, L"VFT_DRV");
    AddType(0x00000004, L"VFT_FONT");
    AddType(0x00000005, L"VFT_VXD");
    AddType(0x00000007, L"VFT_STATIC_LIB");
  }
};
CFileTypeValue g_FileTypeValue;

class CFileDriverSubTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileDriverSubTypeValue()
  {
    AddType(0x00000000, L"VFT2_UNKNOWN");
    AddType(0x00000001, L"VFT2_DRV_PRINTER");
    AddType(0x00000002, L"VFT2_DRV_KEYBOARD");
    AddType(0x00000003, L"VFT2_DRV_LANGUAGE");
    AddType(0x00000004, L"VFT2_DRV_DISPLAY");
    AddType(0x00000005, L"VFT2_DRV_MOUSE");
    AddType(0x00000006, L"VFT2_DRV_NETWORK");
    AddType(0x00000007, L"VFT2_DRV_SYSTEM");
    AddType(0x00000008, L"VFT2_DRV_INSTALLABLE");
    AddType(0x00000009, L"VFT2_DRV_SOUND");
    AddType(0x0000000A, L"VFT2_DRV_COMM");
    AddType(0x0000000B, L"VFT2_DRV_INPUTMETHOD");
  }
};
CFileDriverSubTypeValue g_FileDriverSubTypeValue;

class CFileFontSubTypeValue: public CUINT32FromToUStringMap
{
public:
  CFileFontSubTypeValue()
  {
    AddType(0x00000000, L"VFT2_UNKNOWN");
    AddType(0x00000001, L"VFT2_FONT_RASTER");
    AddType(0x00000002, L"VFT2_FONT_VECTOR");
    AddType(0x00000003, L"VFT2_FONT_TRUETYPE");
  }
};
CFileFontSubTypeValue g_FileFontSubTypeValue;
// -----------------------------
// Load from Binary

static bool GetBoolFromWord(UINT16 word)
{
  switch(word)
  {
    case 0:
      return false;
    case 1:
      return true;
    default:
      throw CDecoderException();
  }
}

void CInfo::LoadFromBinary(CDataInStream &inStream)
{
  UINT16 length = inStream.GetWORD();
  UINT16 valueLength = inStream.GetWORD();
  // int y = sizeof(VS_FIXEDFILEINFO);
  if (valueLength == 0)
    m_FixedFileInfoIsDefined = false;
  else if (valueLength == sizeof(VS_FIXEDFILEINFO))
    m_FixedFileInfoIsDefined = true;
  else
    throw CDecoderException();
  m_IsTextData = GetBoolFromWord(inStream.GetWORD());
  if (m_FixedFileInfoIsDefined)
  {
    int keyLength = (wcslen(kVersionInfoKeyString) + 1) * sizeof(wchar_t);
    if (wcscmp((const wchar_t *)inStream.GetData(), kVersionInfoKeyString) != 0)
      throw CDecoderException();
    inStream.SkipBytes(keyLength);
  }
  inStream.SkipToAlignDWORD();
  memcpy(&m_FixedFileInfo, inStream.GetData(), sizeof(m_FixedFileInfo));
  inStream.SkipBytes(sizeof(m_FixedFileInfo));
  while (inStream.GetBytePosition() < length)
  {
    inStream.SkipToAlignDWORD();
    UInt32 itemEndPosition = inStream.GetBytePosition();
    UInt32 itemLength = inStream.GetWORD();
    itemEndPosition += itemLength;

    UInt32 valueLength = inStream.GetWORD();
    if (valueLength != 0)
      throw CDecoderException();
    bool isTextData = GetBoolFromWord(inStream.GetWORD());;
    UString key = inStream.GetString();
    inStream.SkipToAlignDWORD();
    if (key == kVarFileInfoString)
    {
      CVarFileInfo varFileInfo;
      varFileInfo.m_IsTextData = isTextData;
      inStream.SkipToAlignDWORD();
      while (inStream.GetBytePosition() < itemEndPosition)
      {
        inStream.SkipToAlignDWORD();
        CVar var;
        inStream.GetWORD(); // wLength
        UINT16 valueLength = inStream.GetWORD(); // wLength
        int numItems = valueLength / sizeof(UINT32);
        var.m_IsTextData = GetBoolFromWord(inStream.GetWORD());;
        UString key = inStream.GetString();
        if (key != kTranslationString)
          throw CDecoderException();
        inStream.SkipToAlignDWORD();
        for (int i = 0; i < numItems; i++)
        {
          CLangCodePagePair langCodePagePair;
          UINT32 aDWORD = inStream.GetDWORD();
          langCodePagePair.Lang = LOWORD(aDWORD);
          langCodePagePair.CodePage = HIWORD(aDWORD);
          var.m_LangCodePagePairs.Add(langCodePagePair);
        }
        varFileInfo.m_Vars.Add(var);
      }
      m_VarFileInfoItems.Add(varFileInfo);
      m_StringFileInfoIsUsedVector.Add(false);
    }
    else if (key == kStringFileInfoString)
    {
      CStringFileInfo sFileInfo;
      sFileInfo.m_IsTextData = isTextData;
      inStream.SkipToAlignDWORD();
      while (inStream.GetBytePosition() < itemEndPosition)
      {
        inStream.SkipToAlignDWORD();
        CStringTable sTable;
        UInt32 itemEndPositionofStringTable = inStream.GetBytePosition();
        UInt32 length = inStream.GetWORD();
        itemEndPositionofStringTable += length;
        inStream.GetWORD(); // valueLength
        sTable.m_IsTextData = GetBoolFromWord(inStream.GetWORD());
        sTable.m_Key = inStream.GetString();
        inStream.SkipToAlignDWORD();
        while (inStream.GetBytePosition() < itemEndPositionofStringTable)
        {
          inStream.SkipToAlignDWORD();
          CStringPair sPair;
          inStream.GetWORD(); // Length
          UInt32 valueLength = inStream.GetWORD();
          sPair.m_IsTextData = GetBoolFromWord(inStream.GetWORD());
          sPair.m_Key = inStream.GetString();
          inStream.SkipToAlignDWORD();
          if (valueLength > 0)
            sPair.m_Value = inStream.GetString();
          sTable.m_StringPairs.Add(sPair);
        }
        sFileInfo.m_StringTables.Add(sTable);
      }
      m_StringFileInfoItems.Add(sFileInfo);
      m_StringFileInfoIsUsedVector.Add(true);
    }
    else
      throw CDecoderException();
  }
}


// ----------------------------
// Save to Text

UString GetVersionString(DWORD ms, DWORD ls)
{
  UString s;
  s += UnsignedToUString(HIWORD(ms));
  s += kCharComma;
  s += UnsignedToUString(LOWORD(ms));
  s += kCharComma;
  s += UnsignedToUString(HIWORD(ls));
  s += kCharComma;
  s += UnsignedToUString(LOWORD(ls));
  return s;
}

UString GetOptions(const COptionsCommon &optionsCommon, UINT32 flags)
{
  UString result;
  if (flags == 0)
  {
    if (optionsCommon.GetFirstZeroFlagName(result))
      return result;
    return UnsignedToHexUString(flags);
  }
  UINT32 notUsedFlags;
  UStringVector names;
  optionsCommon.GetOptionNames(flags, names, notUsedFlags);
  for (int i = 0; i < names.Size(); i++)
  {
    if (i != 0)
      result += kCharORWithSpaceDelimeter;
    result += names[i];
  }
  if (notUsedFlags != 0)
  {
    if (names.Size() > 0)
      result += kCharORWithSpaceDelimeter;
    result += UnsignedToHexUString(flags);
  }
  return result;
}

UString GetValue(const CUINT32FromToUStringMap &aMap, UINT32 number)
{
  UString s;
  if (aMap.GetString(number, s))
    return s;
  else
    return UnsignedToUString(number);
}

static UString GetStringWithSpaces(const UString &s, int size)
{
  UString result = s;
  while (result.Length() < size)
    result += kCharSpace;
  result += kCharSpace;
  return result;
}

static UString GetFixedInfoPropertyCaption(const UString &s)
{
  return GetStringWithSpaces(s, 14);
}

void CInfo::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  if (m_FixedFileInfoIsDefined)
  {
    if (m_FixedFileInfo.dwSignature != NFixedFileInfo::kSignature)
      throw CDecoderException();
    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileVersionString);
    outStream += GetVersionString(m_FixedFileInfo.dwFileVersionMS,
        m_FixedFileInfo.dwFileVersionLS);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kProductVersionString);
    outStream += GetVersionString(m_FixedFileInfo.dwProductVersionMS,
        m_FixedFileInfo.dwProductVersionLS);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileFlagsMaskString);
    outStream += UnsignedToHexUString(m_FixedFileInfo.dwFileFlagsMask);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileFlagsString);
    outStream += GetOptions(g_FileFlagsOptions, m_FixedFileInfo.dwFileFlags);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileOSString);
    outStream += GetOptions(g_FileOSOptions, m_FixedFileInfo.dwFileOS);
    AddNewLine(outStream);
  
    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileTypeString);
    outStream += GetValue(g_FileTypeValue, m_FixedFileInfo.dwFileType);
    AddNewLine(outStream);

    outStream += GetFixedInfoPropertyCaption(NFixedFileInfo::kFileSubTypeString);
    switch(m_FixedFileInfo.dwFileType)
    {
      case VFT_DRV:
        outStream += GetValue(g_FileDriverSubTypeValue, m_FixedFileInfo.dwFileSubtype);
        break;
      case VFT_FONT:
        outStream += GetValue(g_FileFontSubTypeValue, m_FixedFileInfo.dwFileSubtype);
        break;
      default:
        outStream += UnsignedToHexUString(m_FixedFileInfo.dwFileSubtype);
    }
    AddNewLine(outStream);
  }
  int stringIndex = 0, varIndex = 0;
  outStream += kBeginString;
  indentMaker.AddIndent();
  for (int i = 0; i < m_StringFileInfoIsUsedVector.Size(); i++)
  {
    AddNewLine(outStream);
    indentMaker.WriteIndent(outStream);
    if (m_StringFileInfoIsUsedVector[i]) // StringTable
    {
      const CStringFileInfo sFileInfo = m_StringFileInfoItems[stringIndex++];
      outStream += kBlockString;
      outStream += kCharSpace;
      outStream += GetQuotedString(kStringFileInfoString);
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kBeginString;
      indentMaker.AddIndent();
      for (int j = 0; j < sFileInfo.m_StringTables.Size(); j++)
      {
        const CStringTable &sTable = sFileInfo.m_StringTables[j];
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kBlockString;
        outStream += kCharSpace;
        outStream += GetQuotedString(sTable.m_Key);
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kBeginString;
        indentMaker.AddIndent();
        for (int k = 0; k < sTable.m_StringPairs.Size(); k++)
        {
          AddNewLine(outStream);
          indentMaker.WriteIndent(outStream);
          outStream += kValueString;
          outStream += kCharSpace;
          const CStringPair &sPair = sTable.m_StringPairs[k];
          outStream += GetStringWithSpaces(GetQuotedString(sPair.m_Key) +
              kCharComma, 20);
          outStream += GetQuotedString(sPair.m_Value);
        }
        indentMaker.DeleteIndent();
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kEndString;
      }
      indentMaker.DeleteIndent();
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kEndString;
    }
    else
    {
      const CVarFileInfo varFileInfo = m_VarFileInfoItems[varIndex++];
      outStream += kBlockString;
      outStream += kCharSpace;
      outStream += GetQuotedString(kVarFileInfoString);
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kBeginString;
      indentMaker.AddIndent();
      for (int j = 0; j < varFileInfo.m_Vars.Size(); j++)
      {
        const CVar &var = varFileInfo.m_Vars[j];
        AddNewLine(outStream);
        indentMaker.WriteIndent(outStream);
        outStream += kValueString;
        outStream += kCharSpace;
        outStream += GetQuotedString(kTranslationString);
        outStream += kPropertyDelimiter;
        for (int k = 0; k < var.m_LangCodePagePairs.Size(); k++)
        {
          const CLangCodePagePair &langCodePagePair = var.m_LangCodePagePairs[k];
          if (k > 0)
          {
            outStream += kPropertyDelimiter;
            outStream += kCharSpace;
          }
          outStream += UnsignedToHexUString(langCodePagePair.Lang);
          outStream += kPropertyDelimiter;
          outStream += UnsignedToUString(langCodePagePair.CodePage);
        }
      }
      indentMaker.DeleteIndent();
      AddNewLine(outStream);
      indentMaker.WriteIndent(outStream);
      outStream += kEndString;
    }
  }
  indentMaker.DeleteIndent();
  AddNewLine(outStream);
  indentMaker.WriteIndent(outStream);
  outStream += kEndString;


  // indentMaker.WriteIndent(outStream);
}


//------------------------------------
// CDecoder

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("txt");

  UINT32 resourceSize = resourceData.Size / 2;
  UINT16 *pData = (UINT16 *)((BYTE *)resourceData.Data);
  CDataInStream inStream(pData, resourceSize);
  UINT16 length = inStream.PeekWORD(0);
  if (length > resourceData.Size)
    throw CDecoderException();

  CIndentMaker indentMaker;
  UString outStream;
  CInfo info;
  info.LoadFromBinary(inStream);
  info.SaveToText(outStream, indentMaker);
  resourceProps.Size = outStream.Length() * sizeof(wchar_t);
  resourceProps.Data.EnsureCapacity(resourceProps.Size);
  memcpy(resourceProps.Data, (LPCWSTR)outStream, resourceProps.Size);
}


}}}
